(* Copyright (C) 2014-15, Daniel Wyckoff *)

(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

Require Import SetUtilities.
Require Import LogicUtilities.
Require Export BoolAlgBasics.
Require Import FieldsOfSets.
Require Import InfiniteOperations.
Require Import TypeUtilities.
Require Import Arith.
Require Import FunctionProperties.
Require Import Equality.
Require Import FunctionalExtensionality.
Require Import IndefiniteDescription.
Require Import FiniteMaps.
Require Import DecidableDec.
Require Import Equality.
Require Import EqDec.
Require Import FiniteOperations.

Section A_sec.
Variable B:Bool_Alg.
Let Bt := Btype (Bc B).
Variable A:(Ensemble Bt). 

Definition SubBtype := {a:Bt | In A a}.

Definition Bplus_sub (x y:SubBtype) := (proj1_sig x) + (proj1_sig y).
Definition Btimes_sub (x y:SubBtype) := (proj1_sig x) * (proj1_sig y).
Definition Bcomp_sub (x:SubBtype) := -(proj1_sig x).

Notation  "x +|' y" := (Bplus_sub x y) (at level 50, left associativity).
Notation  "x *|' y" := (Btimes_sub x y) (at level 40, left associativity).
Notation "-|' x" := (Bcomp_sub x) (at level 30).

Definition Plus_closed_sub : Prop := (forall (x y:SubBtype), 
               In A (x +|' y)).
Definition Times_closed_sub : Prop := (forall (x y:SubBtype), 
               In A (x *|' y)).
Definition One_closed_sub  : Prop := In A 1.
Definition Zero_closed_sub : Prop := In A 0.
Definition Comp_closed_sub : Prop := (forall x:SubBtype, 
               In A (-|' x)).

Definition Plus_closed'' (x y:Bt) :
  Plus_closed_sub -> In A x -> In A y -> In A (x+y).
intros h1 h2 h3.
unfold Plus_closed_sub in h1.
apply h1 with (x := exist (In A) x h2) (y := exist (In A) y h3).
Qed.

Definition Times_closed'' (x y:Bt) :
  Times_closed_sub -> In A x -> In A y -> In A (x*y).
intros h1 h2 h3.
unfold Times_closed_sub in h1.
apply h1 with (x := exist (In A) x h2) (y := exist (In A) y h3).
Qed.

Definition Comp_closed'' (b:Bt) : Comp_closed_sub -> (In A b) ->
  In A (-b).
intros h1 h2.
unfold Comp_closed_sub in h1.
apply h1 with (x := exist (In A) b h2).
Qed.

Record A_closed := {
  P_c : Plus_closed_sub;
  T_c : Times_closed_sub;
  O_c : One_closed_sub;
  Z_c : Zero_closed_sub;
  C_c : Comp_closed_sub}.

Variable Ac : A_closed.

Definition Bc' := 
  Build_Bconst SubBtype
    (Full_set SubBtype)
    (fun (x y:SubBtype) => exist (In A) (x +|' y)
      (P_c Ac x y))
    (fun (x y:SubBtype) => exist (In A) (x *|' y)
      (T_c Ac x y))
    (exist (In A ) 1 (O_c Ac))
    (exist (In A ) 0 (Z_c Ac))
    (fun (x:SubBtype) => (exist (In A) (-|' x)
      (C_c Ac x))).

Section Bc''.


Infix "+|" := (Bplus Bc')  (at level 50, left associativity). 
Infix "*|" := (Btimes Bc') (at level 40, left associativity).
Notation "0|" := (Bzero Bc').
Notation "1|" := (Bone Bc').
Notation "-| x" := ((Bcomp Bc') x) (at level 30).


(*These are just all the basic axioms applied to the subset type SubBtype, derived from the main boolean algebra B *)
Lemma assoc_sum' : forall n m p:SubBtype, n +| (m +| p) = n +| m +| p.
intros n m p.  
red in n.
red in m.
red in p.
inversion n as [x]. 
inversion m as [y]. 
inversion p as [z].
simpl.
unfold Bplus_sub.
simpl.
apply proj1_sig_injective.
simpl. apply assoc_sum.
Qed.

Lemma assoc_prod' : forall n m p:SubBtype, n *| (m *| p) = n *| m *| p.
intros n m p.
red in n.
red in m.
red in p.
inversion n as [x]. 
inversion m as [y]. 
inversion p as [z].
simpl.
unfold Btimes_sub.
simpl.
apply proj1_sig_injective.
simpl. apply assoc_prod.
Qed.

Lemma comm_sum' : forall n m:SubBtype, n +| m = m +| n.
intros n m.
red in n. red in m.
inversion n. inversion m.
simpl.
unfold Bplus_sub.
simpl.
apply proj1_sig_injective.
simpl. apply comm_sum.
Qed.

Lemma comm_prod' : forall n m:SubBtype, n *| m = m *| n.
intros n m.
red in n. red in m.
inversion n. inversion m.
simpl.
unfold Btimes_sub.
simpl.
apply proj1_sig_injective.
simpl. apply comm_prod.
Qed.

Lemma abs_sum'  : forall n m:SubBtype, n +| (n *| m) = n.
intros n m.
red in n. red in m.
inversion n. inversion m.
simpl.
unfold Btimes_sub. unfold Bplus_sub.
simpl.
apply proj1_sig_injective.
simpl. apply abs_sum.
Qed.

Lemma abs_prod':  forall n m:SubBtype, n *| (n +| m) = n.
intros n m.
red in n. red in m.
inversion n. inversion m.
simpl.
unfold Btimes_sub. unfold Bplus_sub.
simpl.
apply proj1_sig_injective.
simpl. apply abs_prod.
Qed.

Lemma dist_sum' : forall n m p:SubBtype, p *| (n +| m) = p *| n + p *| m.
intros n m p.
red in n. red in m. red in p.
inversion n. inversion m. inversion p.
simpl.
unfold Btimes_sub. unfold Bplus_sub.
simpl.
apply proj1_sig_injective.
simpl. apply dist_sum.
Qed.

Lemma dist_prod': forall n m p:SubBtype, p +| (n *| m) = (p +| n) *| (p +| m).
intros n m p.
red in n. red in m. red in p.
inversion n. inversion m. inversion p.
simpl.
unfold Btimes_sub. unfold Bplus_sub.
simpl.
apply proj1_sig_injective.
simpl. apply dist_prod.
Qed.

Lemma comp_sum':  forall n:SubBtype, n +| (-| n) = 1.
intros n.
red in n. inversion n.
simpl.
unfold Bplus_sub. unfold Bcomp_sub.
simpl.
apply proj1_sig_injective.
simpl. apply comp_sum.
Qed.

Lemma comp_prod': forall n:SubBtype, n *| (-| n) = 0.
intros n.
red in n. inversion n.
simpl.
unfold Bplus_sub. unfold Bcomp_sub.
simpl.
apply proj1_sig_injective.
simpl. apply comp_prod.
Qed.

Lemma und_set' : (BS Bc') = Full_set (Btype Bc').
unfold Bc'.
simpl. reflexivity.
Qed.

Definition SubAlg := Build_Bool_Alg Bc' und_set' assoc_sum' 
  assoc_prod' comm_sum' comm_prod' abs_sum' abs_prod' dist_sum'
  dist_prod' comp_sum' comp_prod'.


(*Being closed under 2 operations is sufficient to ensure that a subset is a subalgebra *)
Lemma two_ops_imp_times : Comp_closed_sub -> Plus_closed_sub -> Times_closed_sub.
intros h1 h2. unfold Times_closed_sub.
intros x y.
unfold SubBtype in x.   unfold SubBtype in y.
assert (h3: proj1_sig x * proj1_sig y = -(-(proj1_sig x) + -(proj1_sig y))).
apply times_eq_plus. unfold Btimes_sub. rewrite h3.
apply Comp_closed'' with (b:= (-proj1_sig x + -proj1_sig y)).
apply h1.
apply Plus_closed'' with (x := -proj1_sig x) (y := -proj1_sig y).
apply h2.
apply Comp_closed''.  apply h1. 
apply (proj2_sig x).
apply Comp_closed''.  apply h1.
apply (proj2_sig y).
Qed.

Lemma two_ops_imp_plus : Comp_closed_sub -> Times_closed_sub -> Plus_closed_sub.
intros h1 h2. unfold Plus_closed_sub.
intros x y.
unfold SubBtype in x.   unfold SubBtype in y.
assert (h3: proj1_sig x + proj1_sig y = -(-(proj1_sig x) * -(proj1_sig y))).
apply plus_eq_times. unfold Bplus_sub. rewrite h3.
apply Comp_closed'' with (b:= (-proj1_sig x * -proj1_sig y)).
apply h1.
apply Times_closed'' with (x := -proj1_sig x) (y := -proj1_sig y).
apply h2.
apply Comp_closed''.  apply h1. 
apply (proj2_sig x).
apply Comp_closed''.  apply h1.
apply (proj2_sig y).
Qed.

End Bc''.
End A_sec.

Arguments A_closed [B] _.
Arguments SubAlg [B] _ _.
Arguments Plus_closed_sub [B] _.
Arguments Times_closed_sub [B] _.
Arguments Comp_closed_sub [B] _.
Arguments Zero_closed_sub [B] _.
Arguments One_closed_sub [B] _.

Lemma subalg_functional : 
  forall {B:Bool_Alg} (A A':Ensemble (Btype (Bc B))),
    A = A' ->
    forall (pfa:A_closed A) (pfa':A_closed A'),
      SubAlg A pfa = SubAlg A' pfa'.
intros B A A' h1.
rewrite h1.
intros h2 h3. assert (h4:h2 = h3). apply proof_irrelevance.
subst. reflexivity.
Qed.     
                                                                 

Section Closed.
Variable B:Bool_Alg.
Let Bt := Btype (Bc B).

Lemma plus_closed : 
  forall (S:Ensemble Bt), 
    A_closed S ->
    forall (x y:Bt), In S x -> In S y -> In S (x + y).
intros S h1 x y h2 h3.
pose proof (P_c _ _ h1) as h4.
red in h4.
specialize (h4 (exist _ _ h2) (exist _ _ h3)).
unfold Bplus_sub in h4. simpl in h4. assumption.
Qed.

Lemma times_closed : 
  forall (S:Ensemble Bt), 
    A_closed S ->
    forall (x y:Bt), In S x -> In S y -> In S (x * y).
intros S h1 x y h2 h3.
pose proof (T_c _ _ h1) as h4.
red in h4.
specialize (h4 (exist _ _ h2) (exist _ _ h3)).
unfold Btimes_sub in h4. simpl in h4. assumption.
Qed.

Lemma comp_closed : 
  forall (S:Ensemble Bt), 
    A_closed S ->
    forall x:Bt, In S x -> In S (- x).
intros S h1 x h2.
pose proof (C_c _ _ h1) as h4.
red in h4.
specialize (h4 (exist _ _ h2)).
unfold Bcomp_sub in h4. simpl in h4. assumption.
Qed.

Lemma zero_closed : 
  forall (S:Ensemble Bt), 
    A_closed S -> In S 0.
intros S h1.
pose proof (Z_c _ _ h1) as h4.
red in h4.
assumption.
Qed.

Lemma one_closed : 
  forall (S:Ensemble Bt), 
    A_closed S -> In S 1.
intros S h1.
pose proof (O_c _ _ h1) as h4.
red in h4.
assumption.
Qed.

End Closed.


Section Families_Ind_Families_Subalgebras.
Variable B:Bool_Alg.
Let Bt := Btype (Bc B).
Variable It : Type.
Variable A:Ensemble Bt.

(*Subalgebra -- Indexed representation*)
Record SubAlgI := {  
  sai_A:It->(Ensemble Bt);
  sai_closed:forall (i:It), A_closed (sai_A i);
  sai_B := fun (i:It) =>  SubAlg 
    (sai_A i) (sai_closed i)
  }.

(*Subalgebra -- Family representation*)
Record SubAlgF := {
  saf_A:Family Bt;
  saf_closed : forall (S:Ensemble Bt), In saf_A S -> A_closed S;
  saf_B := fun (S:{S':Ensemble Bt | In saf_A S'}) =>
    SubAlg (proj1_sig S) (saf_closed (proj1_sig S) (proj2_sig S))}.

Variable SI:SubAlgI.
Variable SF:SubAlgF.

Definition int_sai_A := IndexedIntersection (sai_A SI).
Definition int_saf_A := FamilyIntersection (saf_A SF).
 
Lemma Plus_closed_i : Plus_closed_sub int_sai_A.
red.
intros x y.
assert (h1: In int_sai_A (proj1_sig x)).
apply proj2_sig.
assert (h2: In int_sai_A (proj1_sig y)).
apply proj2_sig.
inversion h1. inversion h2.
assert (h3: forall (i:It), Plus_closed_sub (sai_A SI i)).
  intro i.
  apply (P_c _ _ (sai_closed SI i)).
apply indexed_intersection_intro.
intro i.
apply Plus_closed''. 
apply (h3 i). 
apply (H i). 
apply (H1 i).
Qed.

Lemma Plus_closed_F : Plus_closed_sub int_saf_A.
red.
intros x y.
constructor.
intros S h1.
apply Plus_closed''.
apply (P_c _ S (saf_closed SF S h1)).
destruct x as [? h2].
simpl.
inversion h2 as [? h3].
apply (h3 S h1).
destruct y as [? h4].
simpl.
inversion h4 as [? h5].
apply (h5 S h1).
Qed.


(*Dual*)
Lemma Times_closed_i : Times_closed_sub int_sai_A.
red.
intros x y.
assert (h1: In int_sai_A (proj1_sig x)).
apply proj2_sig.
assert (h2: In int_sai_A (proj1_sig y)).
apply proj2_sig.
inversion h1. inversion h2.
assert (h3: forall (i:It), Times_closed_sub (sai_A SI i)).
  intro i.
  apply (T_c _ _ (sai_closed SI i)).
apply indexed_intersection_intro.
intro i.
apply Times_closed''. 
apply (h3 i). 
apply (H i). 
apply (H1 i).
Qed.

Lemma Times_closed_F : Times_closed_sub int_saf_A.
red.
intros x y.
constructor.
intros S h1.
apply Times_closed''.
apply (T_c _ S (saf_closed SF S h1)).
destruct x as [? h2].
simpl.
inversion h2 as [? h3].
apply (h3 S h1).
destruct y as [? h4].
simpl.
inversion h4 as [? h5].
apply (h5 S h1).
Qed.

Lemma Comp_closed_i : Comp_closed_sub int_sai_A.
red.  intro x.
assert (h1: In int_sai_A (proj1_sig x)).
apply proj2_sig.
assert (h2: forall (i:It), Comp_closed_sub (sai_A SI i)).
  intro i.
  apply (C_c _ _ (sai_closed SI i)).
apply indexed_intersection_intro.
intro i.
apply Comp_closed''.
apply (h2 i).
inversion h1.
apply (H i).
Qed.

Lemma Comp_closed_F : Comp_closed_sub int_saf_A.
red.
intros x.
constructor.
intros S h1.
apply Comp_closed''.
apply (C_c _ S (saf_closed SF S h1)).
destruct x as [? h2]. simpl.
inversion h2 as [? h3]. 
apply (h3 S h1).
Qed.

Lemma One_closed_i : One_closed_sub int_sai_A.
red.
apply indexed_intersection_intro.
intro i.
pose proof (O_c _ _ (sai_closed SI i)) as h1.
unfold One_closed_sub in h1.
assumption.
Qed.

Lemma One_closed_F : One_closed_sub int_saf_A.
red.
constructor.
intros S h1.
apply (O_c _ S (saf_closed SF S h1)).
Qed.

Lemma Zero_closed_i : Zero_closed_sub int_sai_A.
red.
apply indexed_intersection_intro.
intro i.
pose proof (Z_c _ _ (sai_closed SI i)) as h1.
unfold Zero_closed_sub in h1.
assumption.
Qed.

Lemma Zero_closed_F : Zero_closed_sub int_saf_A.
red.
constructor.
intros S h1.
apply (Z_c _ S (saf_closed SF S h1)).
Qed.


Lemma int_sai_A_closed : A_closed int_sai_A.
constructor.
apply Plus_closed_i.
apply Times_closed_i.
apply One_closed_i.
apply Zero_closed_i.
apply Comp_closed_i.
Qed.

Lemma int_saf_A_closed : A_closed int_saf_A.
constructor.
apply Plus_closed_F.
apply Times_closed_F.
apply One_closed_F.
apply Zero_closed_F.
apply Comp_closed_F.
Qed.


Definition int_sub_algi := SubAlg _ int_sai_A_closed.
Definition int_sub_algF := SubAlg _ int_saf_A_closed.


(*This defines the family of all subalgebra sets that contain A, as a set*)

Definition S_clo_cont_A := [S:Ensemble Bt | Included A S /\ A_closed S].

(* Ditto as a subalgebra family *)
Definition SA_cont_A : SubAlgF. 
pose S_clo_cont_A as F.
red in F.
assert (h1: forall (S:Ensemble Bt), In F S -> A_closed S).
  intros S h2.
  inversion h2 as [h3].  destruct h3 as [h3l h3r]. assumption.
refine (Build_SubAlgF F h1).
Defined.

Lemma saf_A_eq : saf_A SA_cont_A = S_clo_cont_A.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros S h1.
inversion h1.
constructor; assumption.
(* >= *)
red.
intros S h1.
unfold S_clo_cont_A in h1.
inversion h1.
simpl. constructor; assumption.
Qed.

End Families_Ind_Families_Subalgebras.
Arguments SA_cont_A [B] _.
Arguments int_saf_A [B] _ _.
Arguments int_saf_A_closed [B] _.
Arguments S_clo_cont_A [B] _ _.
Arguments saf_A [B] _ _.
Arguments saf_A_eq [B] _.
Arguments SubAlgF [B].
Arguments saf_closed [B] _ _ _.

Section GenAlg.
Variable B:Bool_Alg.
Let Bt := Btype (Bc B).


Definition Gen_Ens (A:Ensemble Bt) : Ensemble Bt := 
  FamilyIntersection (S_clo_cont_A A).

Lemma closed_gen_ens : forall (E:Ensemble Bt), A_closed (Gen_Ens E).
intro E. unfold Gen_Ens.
constructor.
(* + *)
red.
intros x y.
destruct x as [x h1].
destruct y as [y h2].
destruct h1 as [x h1]. destruct h2 as [y h2].
constructor.
intros S h3. 
pose proof (h1 _ h3) as h4.
pose proof (h2 _ h3) as h5.
unfold Bplus_sub. simpl.
destruct h3 as [h3]. destruct h3 as [h3l h3r].
apply plus_closed; auto.
(* * *)
red.
intros x y.
destruct x as [x h1].
destruct y as [y h2].
destruct h1 as [x h1]. destruct h2 as [y h2].
constructor.
intros S h3. 
pose proof (h1 _ h3) as h4.
pose proof (h2 _ h3) as h5.
unfold Bplus_sub. simpl.
destruct h3 as [h3]. destruct h3 as [h3l h3r].
apply times_closed; auto.
(* 1 *)
red. constructor.
intros S h3.
destruct h3 as [h3]. destruct h3 as [h3l h3r].
apply one_closed; auto.
(* 0 *)
red. constructor.
intros S h3.
destruct h3 as [h3]. destruct h3 as [h3l h3r].
apply zero_closed; auto.
(* - *)
red. intro x.
destruct x as [x h1].
destruct h1 as [x h1].
constructor.
intros S h3.
pose proof (h1 _ h3) as h4.
unfold Bcomp_sub. simpl.
destruct h3 as [h3]. destruct h3 as [h3l h3r].
apply comp_closed; auto.
Qed.

Definition Gen (E:Ensemble Bt) := SubAlg _ (closed_gen_ens E).


Lemma gen_ens_minimal : forall (A S:Ensemble Bt), 
  let F := (S_clo_cont_A A) in In F S -> Included (Gen_Ens A) S.
intros A S F h0.
unfold Gen_Ens.
red.
intros x h2.
inversion h2 as [? h3]. subst.
apply h3.  assumption.
Qed.  

Lemma gen_ens_includes : forall (S:Ensemble Bt), Included S (Gen_Ens S).
intro S.
red.
intros x h1.
constructor.
intros S0 h2.
inversion h2 as [h3].
destruct h3. auto with sets.
Qed.

Lemma gen_minimal'': forall (A S:Ensemble Bt), A_closed S ->
  Included A S -> 
  Included (int_saf_A (SA_cont_A A)) S.
intros A S h1 h2.
unfold int_saf_A.
pose proof (saf_A_eq A) as h3.
rewrite h3.
red.
intros x h4.
inversion h4 as [? h5].
apply h5. constructor. split; assumption.
Qed.
End GenAlg.

Arguments A_closed [B] _.
Arguments Gen_Ens [B] _ _.
Arguments Gen [B] _.


Section Gen_examples.
Variable B:Bool_Alg.
Let Bt := Btype (Bc B).

Example A_closed_two : A_closed (@Couple Bt 0 1).
  constructor.
  (* + *)
  red. intros x y.
  destruct x as [x h2].  destruct y as [y h3].
  unfold Bplus_sub. simpl.
  pose proof (Couple_inv _ _ _ _ h2) as h4.
  pose proof (Couple_inv _ _ _ _ h3) as h5.
  destruct h4 as [h4l | h4r].
    (*h4l*)
    destruct h5 as [h5l | h5r].
      (*h5l*)
      rewrite h4l. rewrite h5l.
      pose proof (@zero_sum B 0) as h6.
      rewrite h6.
      apply Couple_l.
      (*h5r*)
      rewrite h4l. rewrite h5r.
      pose proof (@one_sum B 0) as h6.
      rewrite h6.
      apply Couple_r.
    (*h4r*)
    destruct h5 as [h5l | h5r].
      (*h5l*)
      rewrite h4r. rewrite h5l.
      pose proof (@zero_sum B 1) as h6.
      rewrite h6.
      apply Couple_r.
      (*h5r*)
      rewrite h4r. rewrite h5r.
      pose proof (@one_sum B 1) as h6.
      rewrite h6.
      apply Couple_r.
  (* * *)
  red.  
  intros x y.
  destruct x as [x h1]. destruct y as [y h2].
  unfold Btimes_sub. simpl.
  pose proof (Couple_inv _ _ _ _ h1) as h3.
  pose proof (Couple_inv _ _ _ _ h2) as h4.
  destruct h3 as [h3l | h3r].
  (*h3l*)
    destruct h4 as [h4l | h4r].
    (*h4l*)
    rewrite h3l. rewrite h4l.
    pose proof (@zero_prod B 0) as h5.
    rewrite h5.
    apply Couple_l.
    (*h4r*)
    rewrite h3l. rewrite h4r.
    pose proof (@one_prod B 0) as h5.
    rewrite h5.
    apply Couple_l.
  (*h3r*)
    destruct h4 as [h4l | h4r].
    (*h4l*)
    rewrite h3r. rewrite h4l.
    pose proof (@zero_prod B 1) as h5.
    rewrite h5.
    apply Couple_l.
    (*h4r*)
    rewrite h3r. rewrite h4r.
    pose proof (@one_prod B 1) as h5.
    rewrite h5.
    apply Couple_r.
  (* 1 *)
  red. apply Couple_r.
  (* 0 *)
  red. apply Couple_l.
  (* - *)
  red. intro x.
  destruct x as [x h1].
  unfold Bcomp_sub. simpl.
  pose proof (Couple_inv _ _ _ _ h1) as h2.
  destruct h2 as [h2l | h2r].
  (*h2l*)
  rewrite h2l.
  pose proof (@zero_comp B) as h3.
  rewrite h3.
  apply Couple_r.
  (*h2r*)
  rewrite h2r.
  pose proof (@one_comp B) as h3.
  rewrite h3.
  apply Couple_l.
Qed.


Definition two : Bool_Alg.
pose (@Couple Bt 0 1) as A.
pose proof (A_closed_two) as h1.
refine (SubAlg A h1).
Defined.

Example two_smallest_subalgebra : forall (A:Ensemble Bt), A_closed A -> 
  let C := (Couple 0 1) in Included C A.
intros A h1 C.
red.
intros x h2.
pose proof (Couple_inv _ _ _ _ h2) as h3.
destruct h3 as [h3l | h3r].
rewrite h3l.
apply (Z_c _ _ h1).
rewrite h3r.
apply (O_c _ _ h1).
Qed.

Lemma gen_ens_closed_eq : forall (S:Ensemble Bt), A_closed S -> Gen_Ens S = S.
intros S h1.
apply Extensionality_Ensembles.
red. split.
(* <= *)
apply gen_ens_minimal. 
constructor. split. auto with sets. assumption.
(* >= *)
apply gen_ens_includes.
Qed.

(*"if E is empty and A is not degenerate, then the
subalgebra generated by E is the smallest possible subalgebra of A, namely 2."*)

Example gen_ens_empty : Gen_Ens (Empty_set Bt) = (Couple 0 1).
unfold Gen_Ens. 
apply Extensionality_Ensembles.
red. split.
(* <= *)
apply gen_ens_minimal.
constructor. split; auto with sets.
apply A_closed_two.

(* >= *)
red.
intros x h1.
constructor.
intros S h2.
inversion h2 as [h3].
destruct h3 as [? h4].
pose proof (Couple_inv _ _ _ _ h1) as [h5 | h6].
rewrite h5.
apply (Z_c _ _ h4).
rewrite h6.
apply (O_c _ _ h4).
Qed.

Example gen_empty : Gen (Empty_set Bt) = two.
unfold Gen. unfold two.
generalize (closed_gen_ens B (Empty_set Bt)).
rewrite gen_ens_empty.
intros h1.
assert (h1 = A_closed_two). apply proof_irrelevance.
subst.
reflexivity.
Qed.

End Gen_examples.

(*"A simple but useful remark for subsets E and F of a Boolean algebra A is
that if F is included in E, then the subalgebra generated by F is included (as
a subalgebra) in the subalgebra generated by E."*)

Section More_Gen_Ens.
Variable B:Bool_Alg.
Let Bt := Btype (Bc B).

Lemma gen_ens_preserves_inclusion : forall (S U: Ensemble Bt), Included S U ->
  Included (Gen_Ens S) (Gen_Ens U).
intros S U h1.
unfold Gen_Ens.
red.
intros x h2.
inversion h2 as [? h3].
assert (forall S0 : Ensemble Bt, In (S_clo_cont_A U) S0 -> In S0 x).
  intros S0 h5.
  assert (h6:In (S_clo_cont_A S) S0).
  inversion h5 as [h6].
  destruct h6.
  assert (h7:Included S S0).
    auto with sets.
  constructor. split; assumption.
  apply h3; assumption.
constructor; assumption.
Qed.


(*"The relation of one Boolean algebra being a subalgebra of another is a
partial order on the class of all Boolean algebras."  Technically I don't define
[subalg_of] exactly like that, since I presume an underlying algebra B, but 
my definition is equivalent to the informal definition given in the quotes, 
so I know this definition and lemma are pretty vacuous.*)

Definition subalg_of (A1 A2:{A:Ensemble Bt | A_closed A}) : Prop := 
  (Included (proj1_sig A1) (proj1_sig A2)).

Lemma order_subalg_of : Order _ subalg_of.
constructor.
(*refl*)
red.
intro.
red.
auto with sets.
(*trans*)
red. unfold subalg_of. intros.
auto with sets.
(*antisymm*)
red. unfold subalg_of. intros A1 A2 h1 h2.
cut ((proj1_sig A1) = (proj1_sig A2)).
apply proj1_sig_injective.
auto with sets.
Qed.


(*"The family is said to be directed if any two members B_i
and B_j of the family are always subalgebras of some third member Bk."*)

Definition directed (F:SubAlgF) : Prop := forall (S1 S2:Ensemble Bt), 
  In (saf_A F) S1 -> In (saf_A F) S2 -> (exists (S3:Ensemble Bt),
    In (saf_A F) S3 /\
    Included S1 S3 /\ Included S2 S3).

(*"Lemma 1. The union of a non-empty, directed family of subalgebras is again
a subalgebra."*)

Lemma union_directed_SubAlgf_closed : forall (F:SubAlgF), directed F -> 
  Inhabited (saf_A F) -> A_closed (FamilyUnion (saf_A F)).
intros F h1 h2.
constructor.
(* + *)
red.
intros p q.
destruct p as [p h3]. destruct q as [q h4].
unfold Bplus_sub. simpl.
inversion h3 as [B_i ? h5 h6].
inversion h4 as [B_j ? h7 h8].
unfold directed in h1.
pose proof (h1 _ _ h5 h7) as h9.
elim h9. intros B_k h10.
destruct h10 as [h10 [h11l h11r]].
assert (h12:In B_k p).  auto with sets.
assert (h13:In B_k q).  auto with sets.
pose proof (saf_closed F B_k h10) as h14.
pose proof (P_c _ _ h14) as h15.
red in h15.
pose proof (h15 (exist _ p h12) (exist _ q h13)) as h16.
unfold Bplus_sub in h16.
simpl in h16.
apply family_union_intro with B_k; assumption.
(* * *)
red.
intros p q.
destruct p as [p h3]. destruct q as [q h4].
unfold Btimes_sub. simpl.
inversion h3 as [B_i ? h5 h6].
inversion h4 as [B_j ? h7 h8].
unfold directed in h1.
pose proof (h1 _ _ h5 h7) as h9.
elim h9. intros B_k h10.
destruct h10 as [h10 [h11l h11r]].
assert (h12:In B_k p).  auto with sets.
assert (h13:In B_k q).  auto with sets.
pose proof (saf_closed F B_k h10) as h14.
pose proof (T_c _ _ h14) as h15.
red in h15.
pose proof (h15 (exist _ p h12) (exist _ q h13)) as h16.
unfold Btimes_sub in h16.
simpl in h16.
apply family_union_intro with B_k; assumption.
(* 1 *)
red.
inversion h2 as [S h3].
pose proof (saf_closed F S h3) as h4.
pose proof (O_c _ _ h4) as h5.
red in h5.
apply family_union_intro with S; assumption.
(* 0 *)
red.
inversion h2 as [S h3].
pose proof (saf_closed F S h3) as h4.
pose proof (Z_c _ _ h4) as h5.
red in h5.
apply family_union_intro with S; assumption.
(* - *)
red. intro p.
destruct p as [p h3].
unfold Bcomp_sub. simpl.
inversion h3 as [B_i ? h4 h5].
pose proof (saf_closed F B_i h4) as h6.
pose proof (C_c _ _ h6) as h7.
red in h7.
pose proof (h7 (exist _ p h5)) as h8.
red in h8. unfold Bcomp_sub in h8.
simpl in h8.
apply family_union_intro with B_i; assumption.
Qed.

(* "More precisely, a family {Bi } of subalgebras is a chain if for any two members Bi
and Bj of the family, either Bi is a subalgebra of Bj , or vice versa." *)

Definition chain_subalg (F:@SubAlgF B) : Prop := chain _ (saf_A F).

Lemma chain_impl_directed : forall (F:SubAlgF), chain_subalg F -> directed F.
intros F h1.
red in h1.
red.
intros S1 S2 h2 h3.
red in h1.
pose proof (h1 _ _ h2 h3) as h4.
case h4 as [h4l | h4r].
  (* left *)
  exists S2. split. assumption.
  split. assumption. auto with sets.
  (* right *)
  exists S1. split. assumption.
  split. auto with sets. assumption. 
Qed. 

  
Lemma union_chain_SubAlgf_closed : forall (F:SubAlgF), chain_subalg F -> 
  Inhabited (saf_A F) -> A_closed (FamilyUnion (saf_A F)).
intros F h1 h2.
apply union_directed_SubAlgf_closed.
apply chain_impl_directed; assumption.
assumption.
Qed.

(** "Corollary 1. Let $A$ be a Boolean algebra generated by a set $E$, and for each finite subset $F$ of $E$, let $B_F$ be the subalgebra of $A$ generated by $F$. The family $\{B_F:F \subseteq E \text{ and }  F \text{ is finite}\}$ is directed, and its union is $A$."
[or in my terminology]: 
Assume B is generated by a set E, and for each
finite subset S of E, let A be the subalgebra of B generated by S. The family
{A : A is generated by S and S <= E and S is finite} is directed, and its union is B. *)

Lemma family_included_subs_directed_eq_B : forall (E:Ensemble Bt) (F:SubAlgF), 
  (Full_set Bt) = Gen_Ens E -> (saf_A F) = 
  [A:Ensemble Bt | exists (S:Ensemble Bt), 
    A = Gen_Ens S /\ Included S E /\ Finite S] -> 
  directed F /\ FamilyUnion (saf_A F) = (Full_set Bt).
intros E F h1 h2.
assert (h3:directed F).
  red.
  intros S1' S2' h3 h4.
  rewrite h2.
  rewrite h2 in h3. rewrite h2 in h4.
  inversion h3 as [h5].
  inversion h4 as [h6].
  elim h5. intros S1 h7. destruct h7 as [h7a [h7b h7c]].
  elim h6. intros S2 h8. destruct h8 as [h8a [h8b h8c]].
  assert (h9 :Included S1 (Union S1 S2)). auto with sets.
  assert (h10:Included S2 (Union S1 S2)). auto with sets.
  pose proof (gen_ens_preserves_inclusion S1 (Union S1 S2) h9)  as h11.
  pose proof (gen_ens_preserves_inclusion S2 (Union S1 S2) h10) as h12.
  exists (Gen_Ens (Union S1 S2)).
  constructor. constructor.
  exists (Union S1 S2). split. reflexivity.
  split. auto with sets. apply Union_preserves_Finite;assumption.
  rewrite h7a. rewrite h8a.
  split; assumption.

split. assumption.
(* F's union is Full_set *)
pose proof (union_directed_SubAlgf_closed) as h4.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros x ?. constructor.
(* >= *)
assert (h5:Included E (FamilyUnion (saf_A F))).
  red. intros x h6.
  pose proof (In_singleton _ x) as h7.
  apply family_union_intro with (Gen_Ens (Singleton x)).
  rewrite h2.
  constructor. exists (Singleton x). split. reflexivity.
  split. red. intros x' h8.
  pose proof (Singleton_inv _ _ _ h8) as h9.
  rewrite <- h9. assumption.
  apply Singleton_is_finite.
  constructor. intros S h8.
  inversion h8 as [h9]. destruct h9 as [h9l h9r].
  auto with sets. 
rewrite h1.
apply gen_ens_minimal.
constructor.
split. assumption.
pose proof (Inhabited_intro _ (saf_A F) (Couple 0 1)) as h6.
assert (h7:In (saf_A F) (Couple 0 1)).
  rewrite h2.
  constructor. exists (Empty_set Bt).
  split.
  (*left*) 
  rewrite <- gen_ens_empty. reflexivity.
  (*right*)
  split. auto with sets. constructor.
pose proof (h6 h7) as h8.
pose proof (h4 _ h3 h8). assumption. 
Qed.


(*"For any Boolean algebra A [B in my terminology], 
apply the preceding corollary to the set E = A [B]
to conclude that every Boolean algebra is the (directed) union of its finitely
generated subalgebras."*)

Corollary family_included_subs_directed_eq_B_full_set : 
  forall (F:SubAlgF),
  (saf_A F) = [A:Ensemble Bt | exists (S:Ensemble Bt), 
    A = Gen_Ens S /\ Finite S] -> 
  directed F /\ FamilyUnion (saf_A F) = Full_set Bt.
Proof.
intros F h1.
pose proof (family_included_subs_directed_eq_B (Full_set Bt) F) as h2.
assert (h3:Full_set Bt = Gen_Ens (Full_set Bt)).
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red. intros x ?.
  unfold Gen_Ens.
  constructor.
  intros S h4.
  inversion h4 as [h5].
  destruct h5; auto with sets.
  (* >= *)
  red. intros.  constructor. 
assert (h4:saf_A F = [A : Ensemble Bt
     | exists S : Ensemble Bt,
         A = Gen_Ens S /\ Included S (Full_set Bt) /\ Finite S]).
  rewrite h1.
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red.
  intros S h4.
  inversion h4 as [h5].
  destruct h5 as [E h6].
  destruct h6.
  constructor.
  exists E.
  repeat split; auto with sets.
  (* >= *)
  red.  intros S h4.
  inversion h4 as [h5].
  destruct h5 as [E h6].
  destruct h6 as [? []].
  constructor.
  exists E.
  split; assumption.
apply h2; assumption.
Qed.


End More_Gen_Ens.
 
Require Import ListUtilities.

Section Closed_Finite.
Variable B:Bool_Alg.
Let Bt:= Btype (Bc B).

Lemma plus_list_closed : 
  forall (S:Ensemble Bt) (l:list Bt),
    A_closed S -> Included (list_to_set l) S -> 
    Ensembles.In S (plus_list l).
intros S l.
induction l as [|a l h1].
intros h1 h2. simpl. apply zero_closed. assumption.
intros h2 h3.
simpl. simpl in h3.
assert (h4:Included (list_to_set l) (Add (list_to_set l) a)).
  auto with sets.
assert (h5:Included (list_to_set l) S). auto with sets.
specialize (h1 h2 h5).
apply plus_closed; auto.
pose proof (Add_intro2 _ (list_to_set l) a).
auto with sets.
Qed.

Lemma times_list_closed : 
  forall (S:Ensemble Bt) (l:list Bt),
    A_closed S -> Included (list_to_set l) S -> 
    Ensembles.In S (times_list l).
intros S l.
induction l as [|a l h1].
intros h1 h2. simpl. apply one_closed. assumption.
intros h2 h3.
simpl. simpl in h3.
assert (h4:Included (list_to_set l) (Add (list_to_set l) a)).
  auto with sets.
assert (h5:Included (list_to_set l) S). auto with sets.
specialize (h1 h2 h5).
apply times_closed; auto.
pose proof (Add_intro2 _ (list_to_set l) a).
auto with sets.
Qed.

Lemma plus_set_closed : 
  forall (R S:Ensemble Bt) (pf:Finite R),
    A_closed S -> Included R S ->
    Ensembles.In S (plus_set R pf).
intros R S h1 h2 h3.
pose proof (finite_set_list _  h1) as h4.
destruct h4 as [l h4]. subst.
pose proof (plus_set_compat' _ _ _ h1 (eq_refl _)) as h4.
rewrite h4.
apply plus_list_closed; auto.
Qed.

Lemma times_set_closed : 
  forall (R S:Ensemble Bt) (pf:Finite R),
    A_closed S -> Included R S ->
    Ensembles.In S (times_set R pf).
intros R S h1 h2 h3.
pose proof (finite_set_list _  h1) as h4.
destruct h4 as [l h4]. subst.
pose proof (times_set_compat' _ _ _ h1 (eq_refl _)) as h4.
rewrite h4.
apply times_list_closed; auto.
Qed.

End Closed_Finite.

Section SubCompat.
Variable B:Bool_Alg.
Let Bt:= Btype (Bc B).


Lemma plus_list_sub_compat : 
  forall (A:Ensemble Bt) (pf:A_closed A)
         (l:list (Btype (Bc (SubAlg _ pf)))),
    plus_list (map (@proj1_sig _ _) l) = proj1_sig (plus_list l).
intros S h1 l.
induction l as [|a l h2].
simpl. reflexivity.
simpl. rewrite h2.
unfold Bplus_sub.
reflexivity.
Qed.

Lemma times_list_sub_compat : 
  forall (A:Ensemble Bt) (pf:A_closed A)
         (l:list (Btype (Bc (SubAlg _ pf)))),
    times_list (map (@proj1_sig _ _) l) = proj1_sig (times_list l).
intros S h1 l.
induction l as [|a l h2].
simpl. reflexivity.
simpl. rewrite h2.
unfold Btimes_sub.
reflexivity.
Qed.

Lemma plus_set_sub_compat : forall (A:Ensemble Bt) (pf:A_closed A)
         (E:Ensemble (Btype (Bc (SubAlg _ pf)))) (pfe:Finite E)
         (pfi:Finite (Im E (@proj1_sig _ _))),
    plus_set (Im E (@proj1_sig _ _)) pfi = proj1_sig (plus_set E pfe). 
intros S h1 E h2 h3.
pose proof (finite_set_list _ h2) as h4.
pose proof (finite_set_list _  h3) as h5.
destruct h4 as [l h4].
destruct h5 as [l' h5]. subst.
generalize dependent h3.
rewrite <- map_im_compat.
intro h3.
pose proof plus_set_compat'.
rewrite (plus_set_compat' _ _ _ h3 (eq_refl _)).
rewrite (plus_set_compat' _ _ _ h2 (eq_refl _)).
unfold Bt.
rewrite (plus_list_sub_compat _ h1). 
reflexivity.
Qed.

Lemma times_set_sub_compat : forall (A:Ensemble Bt) (pf:A_closed A)
         (E:Ensemble (Btype (Bc (SubAlg _ pf)))) (pfe:Finite E)
         (pfi:Finite (Im E (@proj1_sig _ _))),
    times_set (Im E (@proj1_sig _ _)) pfi = proj1_sig (times_set E pfe). 
intros S h1 E h2 h3.
pose proof (finite_set_list _ h2) as h4.
pose proof (finite_set_list _  h3) as h5.
destruct h4 as [l h4].
destruct h5 as [l' h5]. subst.
generalize dependent h3.
rewrite <- map_im_compat.
intro h3.
pose proof plus_set_compat'.
rewrite (times_set_compat' _ _ _ h3 (eq_refl _)).
rewrite (times_set_compat' _ _ _ h2 (eq_refl _)).
unfold Bt.
rewrite (times_list_sub_compat _ h1). 
reflexivity.
Qed.


End SubCompat.


Section NormalForm.


Variable A:Bool_Alg.
Let At := Btype (Bc A).


Inductive sign:Set := pls|mns.
Definition signl:list sign := pls::mns::nil.
Definition signe:Ensemble sign := Couple pls mns.

Lemma list_to_set_sign : list_to_set signl = signe.
unfold signl. simpl.
rewrite <- couple_add_add.
unfold signe.
rewrite couple_comm.
reflexivity.
Qed.

Lemma signe_finite : Finite signe.
rewrite <- list_to_set_sign.
apply list_to_set_finite.
Qed.




(*Whereas Givant/Halmos uses "p", I use the Handbook's notation epsilon, but with
  arguments in the same order as G/H's "p" *)
Definition eps(b:At)(sgn:sign) := if sgn then b else -b.
Definition eps'(pr:At*sign) := eps (fst pr) (snd pr).


Lemma eps_eps'_compat : forall x:At, eps x = (fun y:sign => eps' (x, y)).
intro x.
unfold eps. unfold eps'. simpl.  unfold eps.
reflexivity.
Qed.


Definition el_prod {E:Ensemble At} 
           (f:Fin_map E signe mns) : At.
pose (fun i:At => eps  i (f |-> i)) as p.
pose proof (Fin_map_fin_dom f) as h1.
pose proof (finite_image _ _ E p h1) as h2.
refine (times_set _ h2).
Defined.

Definition el_sum {E:Ensemble At} 
           (f:Fin_map E signe mns) : At.
pose (fun i:At => eps  i (f |-> i)) as p.
pose proof (Fin_map_fin_dom f) as h1.
pose proof (finite_image _ _ E p h1) as h2.
refine (plus_set _ h2).
Defined.


Lemma el_prod_le_ai : forall (E:Ensemble At) 
                             (a:Fin_map E signe mns) (i:At),
                        Ensembles.In E i ->
                        le (el_prod a) (eps i (a |-> i)).
intros E a i h1.
unfold el_prod.
apply le_times_set.
apply Im_intro with i; auto.
Qed.

Lemma ai_le_el_sum : forall (E:Ensemble At) 
                             (a:Fin_map E signe mns) (i:At),
                        Ensembles.In E i ->
                        le (eps i (a |-> i))  (el_sum a).
intros E a i h1.
unfold el_sum.
apply le_plus_set.
apply Im_intro with i; auto.
Qed.

Lemma el_prod_disjoint : 
  forall (E:Ensemble At) 
         (a b:Fin_map E signe mns), a <> b ->
             (el_prod a) * (el_prod b) = 0. 
  intros E a b h3.
  pose proof (distinct_fin_maps_differ_at_point _ _ h3) as h4.
  destruct h4 as [i h4].
  destruct h4 as [h4l h4r].
  pose proof (el_prod_le_ai E a i h4l) as h5.
  pose proof (el_prod_le_ai E b i h4l) as h6.
  destruct (a |->i) ; destruct (b |-> i).
  contradict h4r. reflexivity.
  simpl in h5. simpl in h6.
  pose proof (mono_prod _ _ _ _ h5 h6) as h7.
  rewrite comp_prod in h7.
  red in h7.
  rewrite zero_sum in h7. assumption.
  simpl in h5. simpl in h6.
  pose proof (mono_prod _ _ _ _ h5 h6) as h7.
  rewrite (comm_prod _ (- i) i) in h7.
  rewrite comp_prod in h7.
  red in h7.
  rewrite zero_sum in h7. assumption.
  contradict h4r. reflexivity.
Qed.


Lemma el_prod_disjoint' : 
  forall (E:Ensemble At) (X Y:Ensemble (Fin_map E signe mns)),
    Inhabited [x : Fin_map E signe mns * Fin_map E signe mns
            | Ensembles.In (cart_prod X Y) x /\ fst x <> snd x] ->
         (Im
            [x : Fin_map E signe mns * Fin_map E signe mns
            | Ensembles.In (cart_prod X Y) x /\ fst x <> snd x]
            (fun x : Fin_map E signe mns * Fin_map E signe mns =>
             el_prod (fst x) * el_prod (snd x))) = Singleton 0.
intros E X Y h1.
apply Extensionality_Ensembles.
red. split.
red.
intros x h2.
inversion h2 as [x' h3 y h4 h5]. subst. clear h2.
destruct h3 as [h3].
destruct h3 as [h3l h3r].
destruct h3l as [h3l].
destruct h3l as [h3a h3b].
pose proof (el_prod_disjoint _ (fst x') (snd x') h3r) as h4.
rewrite h4. constructor.
destruct h1 as [pr h1].
destruct h1 as [h1].
destruct h1 as [h1l h1r].
pose proof (el_prod_disjoint _ (fst pr) (snd pr) h1r) as h2.  
red.
intros x h1. destruct h1; subst.
rewrite <- h2.
apply Im_intro with pr.
constructor; auto.
reflexivity.
Qed.

Lemma plus_set_el_prod_disjoint : 
  forall (E:Ensemble At) (X Y:Ensemble (Fin_map E signe mns))
         (pf:Finite [x : Fin_map E signe mns * Fin_map E signe mns
            | Ensembles.In (cart_prod X Y) x /\ fst x <> snd x]),
         plus_set (Im
            [x : Fin_map E signe mns * Fin_map E signe mns
            | Ensembles.In (cart_prod X Y) x /\ fst x <> snd x]
            (fun x : Fin_map E signe mns * Fin_map E signe mns =>
             el_prod (fst x) * el_prod (snd x))) (finite_image _ _ _ _ pf) = 0.
intros E X Y h1.
destruct (classic (Inhabited ([x : Fin_map E signe mns * Fin_map E signe mns
         | Ensembles.In (cart_prod X Y) x /\ fst x <> snd x]))) as [h2 | h3].
pose proof (@el_prod_disjoint').
pose proof (el_prod_disjoint' _ X Y h2) as h3.
generalize  (finite_image (Fin_map E signe mns * Fin_map E signe mns) 
        (Btype (Bc A))
        [x : Fin_map E signe mns * Fin_map E signe mns
        | Ensembles.In (cart_prod X Y) x /\ fst x <> snd x]
        (fun x : Fin_map E signe mns * Fin_map E signe mns =>
         el_prod (fst x) * el_prod (snd x)) h1).
rewrite h3.
intro. apply plus_set_sing'.
pose proof (not_inhabited_empty _ h3) as h4.
pose proof image_empty.
pose proof (image_empty _ _ (fun x : Fin_map E signe mns * Fin_map E signe mns =>
         el_prod (fst x) * el_prod (snd x))) as h5.
rewrite <- h4 in h5.
pose proof (plus_set_functional _ _  _ (finite_image (Fin_map E signe mns * Fin_map E signe mns) 
        (Btype (Bc A))
        [x : Fin_map E signe mns * Fin_map E signe mns
        | Ensembles.In (cart_prod X Y) x /\ fst x <> snd x]
        (fun x : Fin_map E signe mns * Fin_map E signe mns =>
         el_prod (fst x) * el_prod (snd x)) h1) (Empty_is_finite _) h5) as h6.
rewrite plus_set_empty in h6.
assumption.
Qed.

Lemma plus_set_el_prod_disjoint' : 
  forall (E:Ensemble At) (X Y:Ensemble (Fin_map E signe mns))
         (pf:Finite [x : Fin_map E signe mns * Fin_map E signe mns
            | Ensembles.In (cart_prod X Y) x /\ fst x <> snd x])
         (pfi:Finite (Im
            [x : Fin_map E signe mns * Fin_map E signe mns
            | Ensembles.In (cart_prod X Y) x /\ fst x <> snd x]
            (fun x : Fin_map E signe mns * Fin_map E signe mns =>
             el_prod (fst x) * el_prod (snd x)))),
         plus_set (Im
            [x : Fin_map E signe mns * Fin_map E signe mns
            | Ensembles.In (cart_prod X Y) x /\ fst x <> snd x]
            (fun x : Fin_map E signe mns * Fin_map E signe mns =>
             el_prod (fst x) * el_prod (snd x))) pfi = 0.
intros E X Y h1 h2.
pose proof (plus_set_el_prod_disjoint _ X Y h1) as h3.
rewrite <- h3.
apply plus_set_functional. reflexivity.
Qed.


Lemma eps_covers : 
  forall (E:Ensemble At) (pf:Finite E) (x:At),
    plus_set (Im signe (eps x)) (finite_image _ _ signe (eps x) (signe_finite)) = 1.
intros E h1 x.
assert (h2:list_to_set ((eps x pls) :: (- (eps x pls)) :: nil) =
           (Im signe (eps x))).
  apply Extensionality_Ensembles.
  red. split.
  red.
  intros y h3.
  rewrite <- list_to_set_in_iff in h3.
  simpl in h3.
  destruct h3 as [h3a | [h3b | h3c]].
  subst.
  apply Im_intro with pls. constructor. simpl. reflexivity.
  apply Im_intro with mns. constructor. simpl. subst. reflexivity.
  contradiction.
  red.
  intros y h3.
  destruct h3 as [y h3 z]. subst.
  unfold signe in h3.
  destruct h3.
  simpl. right. constructor. simpl. left. right. constructor.
rewrite (plus_set_compat' _ _ (eps x pls :: - eps x pls :: nil)).
simpl.
rewrite zero_sum.
rewrite comp_sum. reflexivity.
unfold At. unfold At in h2.
rewrite h2.
reflexivity.
Qed.

 
Lemma el_prod_covers : 
  forall (E:Ensemble At) (pf:Finite E),
    plus_set (Im (Full_set (Fin_map E signe mns)) el_prod) 
    (finite_image _ _ _ _ (finite_fin_maps _ _ mns pf signe_finite)) = 1.
intros E h1.
assert (h2:times_plus_fin_pair_map1
      _ h1 (fun_to_fin_map _ 0 (cart_prod_fin _ _ h1 signe_finite) eps') = 1).
    unfold times_plus_fin_pair_map1.
  unfold fin_map_times.
generalize ((finite_image At (Btype (Bc A)) E
        (fin_map_app
           (plus_fin_pair_map1 A h1
              (fun_to_fin_map (cart_prod E signe) 0
                 (cart_prod_fin E signe h1 signe_finite) eps')))
        (Fin_map_fin_dom
           (plus_fin_pair_map1 A h1
              (fun_to_fin_map (cart_prod E signe) 0
                 (cart_prod_fin E signe h1 signe_finite) eps'))))).
unfold plus_fin_pair_map1.
rewrite im_fin_map_app_undoes_fun_to_fin_map.
intros h2. 
assert (h3: Im E (fun x : At =>
         plus_set
           (im1
              (fun_to_fin_map (cart_prod E signe) 0
                 (cart_prod_fin E signe h1 signe_finite) eps') x)
           (im1_fin
              (fun_to_fin_map (cart_prod E signe) 0
                 (cart_prod_fin E signe h1 signe_finite) eps') x)) =
           Im E (fun x:At => plus_set (Im signe (fun y => (eps' (x, y)))) (finite_image _ _ _ _ signe_finite))).
apply Extensionality_Ensembles.
red. split.
red. intros b h3.
destruct h3 as [a h4 b h5].
apply Im_intro with a. assumption.
rewrite h5.
apply plus_set_functional.
apply im1_fun_to_fin_map; auto.
red. intros b h3.
destruct h3 as [a h4 b h5].
apply Im_intro with a. assumption.
rewrite h5.
apply plus_set_functional.
symmetry. apply im1_fun_to_fin_map; auto.
revert h2.
rewrite h3. clear h3.
induction h1 as [|E h2 h3 e h4].
intros h2.
pose proof (image_empty _ _  (fun x : At =>
         plus_set (Im signe (fun y : sign => eps' (x, y)))
           (finite_image sign (Btype (Bc A)) signe
              (fun y : sign => eps' (x, y)) signe_finite))) as h3.
revert h2.
rewrite h3.
intros h2. apply times_set_empty'.
intro h5.
rewrite <- times_set_sing.
apply times_set_functional.
apply Extensionality_Ensembles.
red. split.
red.
intros a h6.
destruct h6 as [x h6 a h7].
rewrite <- (eps_eps'_compat x) in h7.
pose proof eps_covers.
pose proof (eps_covers _ h2 x) as h8.
unfold At in h8. unfold At in h7.
rewrite h8 in h7.
subst. constructor.
red. intros x h1.
destruct h1; subst.
apply Im_intro with e. right. constructor. 
rewrite <- (eps_eps'_compat e).
rewrite <- (eps_covers _ h2 e).
reflexivity.
rewrite <- h2.
rewrite (complete_dist_times_plus1 _ _ _ mns).
unfold plus_times_fun_all_maps1.
apply plus_set_functional.
f_equal. 
apply functional_extensionality.
intro F.
unfold el_prod. unfold times_fun_fin_map1. unfold fin_map_times.
apply times_set_functional.
apply Extensionality_Ensembles.
red. split.
intro  y.
intros h3.
destruct h3 as [e h4 y h5].
apply Im_intro with e. assumption.
rewrite fun_to_fin_map_compat. rewrite fun_to_fin_map_compat.
unfold eps'. simpl. assumption.
split. simpl.
split; auto.
apply fin_map_app_in. assumption. assumption.
red. 
intros y h3.
destruct h3 as [e h4 y h5].
rewrite fun_to_fin_map_compat in h5. rewrite fun_to_fin_map_compat in h5.
apply Im_intro with e. assumption. unfold eps' in h5. simpl in h5.
assumption.
split. simpl. split; auto. apply fin_map_app_in. assumption.
assumption.
Qed.

Lemma el_prod_covers' : 
  forall (E:Ensemble At) (pf:Finite E)
         (pf:Finite (Im (Full_set (Fin_map E signe mns)) el_prod)),
    plus_set (Im (Full_set (Fin_map E signe mns)) el_prod) pf = 1.
intros E h1 h2.
pose proof (el_prod_covers _ h1) as h3.
rewrite <- h3.
apply plus_set_functional. reflexivity.
Qed.


Definition non_zero_el_prod_maps (E:Ensemble At) :=
  [a:Fin_map E signe mns | el_prod a <> 0].

Lemma non_zero_el_prod_maps_fin : 
  forall (E:Ensemble At), Finite E ->
                          Finite (non_zero_el_prod_maps E).
intros E h1.
assert (h2:Included (non_zero_el_prod_maps E) (Full_set (Fin_map E signe mns))). red. intros; constructor.
pose proof (finite_fin_maps _ _ mns h1 signe_finite) as h3.
apply (Finite_downward_closed _ _  h3 _ h2).
Qed.

Definition plus_subset_non_zero_el_prod_maps (E:Ensemble At) (pfe:Finite E) (S:Ensemble (Fin_map E signe mns))
           (pfi:Included S (non_zero_el_prod_maps E)) : At :=
  plus_set (Im S el_prod) (finite_image _ _ _ el_prod (Finite_downward_closed _ _ (non_zero_el_prod_maps_fin _ pfe) _ pfi)).

Lemma inclusion_plus_subset_non_zero_el_prod_maps_preserves_int : 
  forall (E:Ensemble At) (pfe:Finite E) (X Y:Ensemble (Fin_map E signe mns)),
    Included X (non_zero_el_prod_maps E) ->
    Included Y (non_zero_el_prod_maps E) ->
    Included (Intersection X Y) (non_zero_el_prod_maps E).
intros E h1 X Y h2 h3.
red.
intros a h4.
destruct h4 as [a h4l h4r].
constructor.
red in h3.
specialize (h3 _ h4r).
destruct h3.
assumption.
Qed.

Lemma inclusion_plus_subset_non_zero_el_prod_maps_preserves_union : 
  forall (E:Ensemble At) (pfe:Finite E) (X Y:Ensemble (Fin_map E signe mns)),
    Included X (non_zero_el_prod_maps E) ->
    Included Y (non_zero_el_prod_maps E) ->
    Included (Union X Y) (non_zero_el_prod_maps E).
intros E h1 X Y h2 h3. red.
intros a h4.
destruct h4 as [a h4l | a h4r].
red in h2. specialize (h2 _ h4l). assumption.
red in h3. specialize (h3 _ h4r). assumption.
Qed.


Lemma inclusion_plus_subset_non_zero_el_prod_maps_preserves_comp : 
  forall (E:Ensemble At) (pfe:Finite E) (X:Ensemble (Fin_map E signe mns)),
    Included X (non_zero_el_prod_maps E) ->
    Included (Setminus (non_zero_el_prod_maps E) X) (non_zero_el_prod_maps E).
intros E h1 X h2. red.
intros a h4.
destruct h4 as [h4l].
assumption.
Qed.

Lemma inclusion_plus_subset_non_zero_el_prod_maps_preserves_setminus : 
  forall (E:Ensemble At) (pfe:Finite E) (X Y:Ensemble (Fin_map E signe mns)),
    Included X (non_zero_el_prod_maps E) ->
    Included (Setminus X Y) (non_zero_el_prod_maps E).
intros E h1 X Y h2. red.
intros a h3.
destruct h3 as [h3a h3b].
auto with sets.
Qed.

Lemma inclusion_plus_subset_non_zero_el_prod_maps_preserves_symdiff_full : 
  forall (E:Ensemble At) (pfe:Finite E) (X Y:Ensemble (Fin_map E signe mns)),
    Included X (non_zero_el_prod_maps E) ->
    Included Y (non_zero_el_prod_maps E) ->
    Included (Symdiff_full (non_zero_el_prod_maps E) X Y) (non_zero_el_prod_maps E).
intros E h1 X Y h2 h3. red.
intros a h4.
destruct h4 as [a h4l | a h4r].
destruct h4l as [a h4a h4b].  auto with sets.
destruct h4r as [a h4a h4b]. auto with sets.
Qed.



Lemma plus_subset_non_zero_el_prod_maps_empty :
  forall (E:Ensemble At) (pfe:Finite E),
    plus_subset_non_zero_el_prod_maps E pfe _ (empty_inclusion _) = 0.
intros. unfold plus_subset_non_zero_el_prod_maps.
generalize ((finite_image (Fin_map E signe mns) At (Empty_set (Fin_map E signe mns))
        el_prod
        (Finite_downward_closed (Fin_map E signe mns)
           (non_zero_el_prod_maps E) (non_zero_el_prod_maps_fin E pfe)
           (Empty_set (Fin_map E signe mns))
           (empty_inclusion (non_zero_el_prod_maps E))))).
pose proof (image_empty (Fin_map E signe mns) _ el_prod) as h1.
rewrite h1.
intro. apply plus_set_empty'.
Qed.

Lemma le_plus_subset_non_zero_el_prod_maps : 
  forall (E:Ensemble At) (pfe:Finite E) (X:Ensemble (Fin_map E signe mns)) 
         (pf:Included X (non_zero_el_prod_maps E)) (a:Fin_map E signe mns), 
    Ensembles.In X a -> le (el_prod a) (plus_subset_non_zero_el_prod_maps E pfe _ pf).
intros E h1 X h2 a h3.
unfold plus_subset_non_zero_el_prod_maps.
apply le_plus_set.
apply Im_intro with a; auto.
Qed.

Lemma plus_subset_non_zero_el_prod_maps_empty_rev :
  forall (E:Ensemble At) (pfe:Finite E) (X:Ensemble (Fin_map E signe mns))
         (pf:Included X (non_zero_el_prod_maps E)),
         plus_subset_non_zero_el_prod_maps E pfe _ pf = 0 -> X = Empty_set _.
intros E h1 X h2 h3.
apply NNPP.
intros h4.
pose proof (not_empty_Inhabited _ _ h4) as h5.
destruct h5 as [a h5].
unfold non_zero_el_prod_maps in h2.
assert (h6:Ensembles.In [a:Fin_map E signe mns | el_prod a <> 0] a). auto with sets.
destruct h6 as [h6].
pose proof (le_plus_subset_non_zero_el_prod_maps _ h1 _ h2 _ h5) as h7.
contradict h6.
rewrite h3 in h7.
apply le_x_0.
assumption.
Qed.
                                                                                                  
Lemma plus_set_el_prod_zero : forall (E:Ensemble At) (pf:Finite _), plus_set (Im [x : Fin_map E signe mns | el_prod x = 0] el_prod) pf = 0.
intros E h1.
destruct (classic (Im [x : Fin_map E signe mns | el_prod x = 0] el_prod = Empty_set _)) as [h2 | h3].
generalize h1.
rewrite h2. intro. apply plus_set_empty'.
pose proof (not_empty_Inhabited _ _ h3) as h4.
rewrite <- plus_set_sing.
apply plus_set_functional.
apply Extensionality_Ensembles.
red. split. red. intros x h15.
inversion h15 as [a h16 y]. subst.
inversion h15 as [b h17 z]. subst.
destruct h17 as [h17]. rewrite <- H in h17.
rewrite h17. constructor.
red. intros x h15.
destruct h15. subst.
destruct h4 as [a h4].
inversion h4. subst.
inversion h4. subst.
inversion H0. 
rewrite H2 in H1.
rewrite <- H1.
apply Im_intro with x. constructor. reflexivity. reflexivity.
Qed.


Lemma plus_subset_non_zero_el_prod_maps_full : 
  forall (E:Ensemble At) (pfe:Finite E),
    plus_subset_non_zero_el_prod_maps E pfe _ (inclusion_reflexive _) = 1.
intros E h1.
unfold plus_subset_non_zero_el_prod_maps.
unfold non_zero_el_prod_maps.
pose proof (excl_middle_sat (fun a:(Fin_map E signe mns) => el_prod a = 0)) as h2.
pose proof f_equal.
pose proof (feq_im _ _ el_prod h2) as h3.
simpl in h3. 
rewrite image_union in h3.
pose proof (finite_fin_maps _ _ mns h1 signe_finite) as h4. 
pose proof (finite_image _ _ _ el_prod h4) as h5.
pose proof h5 as h6.
rewrite h3 in h6.
pose proof f_equal.
pose proof (plus_set_functional _ _ _ h5 h6 h3) as h7.
pose proof (el_prod_covers' _ h1 h5) as h8.
rewrite h8 in h7. symmetry in h7.
pose proof (plus_set_union).
assert (h9:Included (Im [x : Fin_map E signe mns | el_prod x = 0] el_prod)  (Im (Full_set (Fin_map E signe mns)) el_prod)).
rewrite h3. auto with sets.
assert (h10:Included (Im [x : Fin_map E signe mns | el_prod x <> 0] el_prod)  (Im (Full_set (Fin_map E signe mns)) el_prod)).
rewrite h3. auto with sets.
pose proof (Finite_downward_closed).
pose proof (Finite_downward_closed _ _ h5 _ h9) as h11.
pose proof (Finite_downward_closed _ _ h5 _ h10) as h12.
pose proof (plus_set_union _ _ _ h11 h12) as h13.
assert (h14:(Union_preserves_Finite (Btype (Bc A))
             (Im [x : Fin_map E signe mns | el_prod x = 0] el_prod)
             (Im [x : Fin_map E signe mns | el_prod x <> 0] el_prod) h11 h12) = h6). apply proof_irrelevance.
rewrite h14 in h13.
unfold At in h7. unfold At in h13.
rewrite h7 in h13. 
pose proof plus_subset_non_zero_el_prod_maps_empty.
pose proof (plus_set_el_prod_zero _ h11) as h15.
unfold At in h15. unfold At in h13. rewrite h15 in h13.
rewrite comm_sum in h13.
rewrite zero_sum in h13.
rewrite h13.
apply plus_set_functional. reflexivity.
Qed.



Lemma plus_subset_non_zero_el_prod_maps_int :
  forall (E:Ensemble At) (pfe:Finite E) 
         (X Y:Ensemble (Fin_map E signe mns))
         (pfx:Included X (non_zero_el_prod_maps E))
         (pfy:Included Y (non_zero_el_prod_maps E)),
         (plus_subset_non_zero_el_prod_maps E pfe _ pfx) *
         (plus_subset_non_zero_el_prod_maps E pfe _ pfy) =
         plus_subset_non_zero_el_prod_maps E pfe (Intersection X Y) (inclusion_plus_subset_non_zero_el_prod_maps_preserves_int _ pfe X Y pfx pfy).
intros E h1 X Y h2 h3.
unfold plus_subset_non_zero_el_prod_maps.
rewrite dist_set_plus2.
pose proof (cart_prod_im X Y el_prod el_prod) as h4.
pose proof (excl_middle_sat' (cart_prod X Y) (fun pr => fst pr = snd pr)) as h5.
rewrite h5 in h4. clear h5.
rewrite image_union in h4.
pose proof (feq_im _ _ (fun p : Btype (Bc A) * Btype (Bc A) => fst p * snd p) h4) as h5.
rewrite image_union in h5.
do 2 rewrite im_im in h5. simpl in h5.
pose proof (plus_set_functional).
pose proof (finite_fin_map_ens X h1 signe_finite) as h6.
pose proof (finite_fin_map_ens Y h1 signe_finite) as h7.
pose proof (finite_image _ _ _ el_prod h6) as h8.
pose proof (finite_image _ _ _ el_prod h7) as h9.
pose proof (cart_prod_fin _ _ h8 h9) as h10.
pose proof (finite_image).
pose proof (finite_image _ _ _  (fun p : Btype (Bc A) * Btype (Bc A) => fst p * snd p) h10) as h11.
rewrite h5 in h11.
pose proof (plus_set_functional _ _ _ (finite_image (Btype (Bc A) * Btype (Bc A)) (Btype (Bc A))
        (cart_prod (Im X el_prod) (Im Y el_prod))
        (fun p : Btype (Bc A) * Btype (Bc A) => fst p * snd p)
        (cart_prod_fin (Im X el_prod) (Im Y el_prod)
           (finite_image (Fin_map E signe mns) At X el_prod
              (Finite_downward_closed (Fin_map E signe mns)
                 (non_zero_el_prod_maps E) (non_zero_el_prod_maps_fin E h1) X
                 h2))
           (finite_image (Fin_map E signe mns) At Y el_prod
              (Finite_downward_closed (Fin_map E signe mns)
                 (non_zero_el_prod_maps E) (non_zero_el_prod_maps_fin E h1) Y
                 h3)))) h11 h5) as h12.
assert (h13:Included (Im
                [x : Fin_map E signe mns * Fin_map E signe mns
                | Ensembles.In (cart_prod X Y) x /\ fst x = snd x]
                (fun x : Fin_map E signe mns * Fin_map E signe mns =>
                 el_prod (fst x) * el_prod (snd x)))
  (Union
             (Im
                [x : Fin_map E signe mns * Fin_map E signe mns
                | Ensembles.In (cart_prod X Y) x /\ fst x = snd x]
                (fun x : Fin_map E signe mns * Fin_map E signe mns =>
                 el_prod (fst x) * el_prod (snd x)))
             (Im
                [x : Fin_map E signe mns * Fin_map E signe mns
                | Ensembles.In (cart_prod X Y) x /\ fst x <> snd x]
                (fun x : Fin_map E signe mns * Fin_map E signe mns =>
                 el_prod (fst x) * el_prod (snd x))))). auto with sets.

assert (h14:Included (Im
                [x : Fin_map E signe mns * Fin_map E signe mns
                | Ensembles.In (cart_prod X Y) x /\ fst x <> snd x]
                (fun x : Fin_map E signe mns * Fin_map E signe mns =>
                 el_prod (fst x) * el_prod (snd x)))
  (Union
             (Im
                [x : Fin_map E signe mns * Fin_map E signe mns
                | Ensembles.In (cart_prod X Y) x /\ fst x = snd x]
                (fun x : Fin_map E signe mns * Fin_map E signe mns =>
                 el_prod (fst x) * el_prod (snd x)))
             (Im
                [x : Fin_map E signe mns * Fin_map E signe mns
                | Ensembles.In (cart_prod X Y) x /\ fst x <> snd x]
                (fun x : Fin_map E signe mns * Fin_map E signe mns =>
                 el_prod (fst x) * el_prod (snd x))))). auto with sets.
pose proof (Finite_downward_closed _ _ h11 _ h13) as h15. clear h13.
pose proof (Finite_downward_closed _ _ h11 _ h14) as h16. clear h14.
rewrite (plus_set_union' _ _ _ h15 h16 h11) in h12.
pose proof (finite_fin_map_squared_ens [x : Fin_map E signe mns * Fin_map E signe mns
            | Ensembles.In (cart_prod X Y) x /\ fst x <> snd x] h1 signe_finite) as h13.
rewrite (plus_set_el_prod_disjoint' _ X Y h13 h16) in h12.
rewrite zero_sum in h12.
unfold At. unfold At in h12.
rewrite h12.
apply plus_set_functional.
(* maybe move this part to Set_Utilities somehow*)
apply Extensionality_Ensembles.
red. split.
red.
intros x h17.
inversion h17 as [a h18 b h19 h20]. subst.
destruct h18 as [h18].
destruct h18 as [h18l h18r].
destruct h18l as [h18l].
destruct h18l as [h18a h18b].
rewrite h18r.
rewrite idem_prod.
rewrite h18r in h18a.
apply Im_intro with (snd a).
split; auto. reflexivity.
red.
intros x h17.
inversion h17 as [a h18 b h19 h20]. subst.
destruct h18 as [a h18l h18r].
apply Im_intro with (a, a).
constructor.
simpl.
split; auto. split; auto.
simpl. rewrite idem_prod. reflexivity.
Qed.


Lemma plus_subset_non_zero_el_prod_maps_union :
  forall (E:Ensemble At) (pfe:Finite E) 
         (X Y:Ensemble (Fin_map E signe mns))
         (pfx:Included X (non_zero_el_prod_maps E))
         (pfy:Included Y (non_zero_el_prod_maps E)),
         (plus_subset_non_zero_el_prod_maps E pfe _ pfx) +
         (plus_subset_non_zero_el_prod_maps E pfe _ pfy) =
         plus_subset_non_zero_el_prod_maps E pfe (Union X Y) (inclusion_plus_subset_non_zero_el_prod_maps_preserves_union _ pfe X Y pfx pfy).
intros E h1 X Y h2 h3. 
unfold plus_subset_non_zero_el_prod_maps.
generalize (finite_image (Fin_map E signe mns) At (Union X Y) el_prod
        (Finite_downward_closed (Fin_map E signe mns)
           (non_zero_el_prod_maps E) (non_zero_el_prod_maps_fin E h1)
           (Union X Y)
           (inclusion_plus_subset_non_zero_el_prod_maps_preserves_union E h1
              X Y h2 h3))).
rewrite <- plus_set_union.
intro h4.
apply plus_set_functional.
rewrite image_union.
reflexivity.
Qed.

Lemma plus_subset_non_zero_el_prod_maps_comp :
  forall (E:Ensemble At) (pfe:Finite E) 
         (X:Ensemble (Fin_map E signe mns))
         (pfx:Included X (non_zero_el_prod_maps E)),
         - (plus_subset_non_zero_el_prod_maps E pfe _ pfx) =
         plus_subset_non_zero_el_prod_maps E pfe _ (inclusion_plus_subset_non_zero_el_prod_maps_preserves_comp _ pfe X pfx).
intros E h1 X h2.
assert (h3:(plus_subset_non_zero_el_prod_maps E h1 X h2) *
   plus_subset_non_zero_el_prod_maps E h1
     (Setminus (non_zero_el_prod_maps E) X)
     (inclusion_plus_subset_non_zero_el_prod_maps_preserves_comp E h1 X h2) = 0).
  rewrite plus_subset_non_zero_el_prod_maps_int.
  rewrite <- (plus_subset_non_zero_el_prod_maps_empty E h1).
  unfold plus_subset_non_zero_el_prod_maps.
  apply plus_set_functional.
  apply feq_im.
  apply int_setminus.
assert (h4:(plus_subset_non_zero_el_prod_maps E h1 X h2) +
   plus_subset_non_zero_el_prod_maps E h1
     (Setminus (non_zero_el_prod_maps E) X)
     (inclusion_plus_subset_non_zero_el_prod_maps_preserves_comp E h1 X h2) = 1).
  rewrite plus_subset_non_zero_el_prod_maps_union.
  rewrite <- (plus_subset_non_zero_el_prod_maps_full E h1).
  unfold plus_subset_non_zero_el_prod_maps.
  apply plus_set_functional. apply feq_im.
  apply union_setminus_inc. assumption.
pose proof (comp_char _ _ h3 h4) as h5.
symmetry.
assumption.
Qed.

Lemma plus_subset_non_zero_el_prod_maps_setminus :
  forall (E:Ensemble At) (pfe:Finite E) 
         (X Y:Ensemble (Fin_map E signe mns))
         (pfx:Included X (non_zero_el_prod_maps E))
         (pfy:Included Y (non_zero_el_prod_maps E)),
         (plus_subset_non_zero_el_prod_maps E pfe _ pfx) *
         - (plus_subset_non_zero_el_prod_maps E pfe _ pfy) =
         plus_subset_non_zero_el_prod_maps E pfe 
                                           (Intersection X (Setminus (non_zero_el_prod_maps E) Y))
                                                         (inclusion_plus_subset_non_zero_el_prod_maps_preserves_int _ pfe X _ pfx
                                                         (inclusion_plus_subset_non_zero_el_prod_maps_preserves_comp _ pfe Y pfy)).
intros E h1 X Y h2 h3. 
rewrite <- plus_subset_non_zero_el_prod_maps_int.
rewrite <- plus_subset_non_zero_el_prod_maps_comp.
reflexivity.
Qed.



Lemma plus_subset_non_zero_el_prod_maps_symdiff_full :
  forall (E:Ensemble At) (pfe:Finite E) 
         (X Y:Ensemble (Fin_map E signe mns))
         (pfx:Included X (non_zero_el_prod_maps E))
         (pfy:Included Y (non_zero_el_prod_maps E)),
         (plus_subset_non_zero_el_prod_maps E pfe _ pfx) /_\
         (plus_subset_non_zero_el_prod_maps E pfe _ pfy) =
         plus_subset_non_zero_el_prod_maps E pfe (Symdiff_full (non_zero_el_prod_maps E) X Y) (inclusion_plus_subset_non_zero_el_prod_maps_preserves_symdiff_full _ pfe X Y pfx pfy).
intros E h1 X Y h2 h3. 
unfold sym_diff.
do 2 rewrite plus_subset_non_zero_el_prod_maps_comp.
do 2 rewrite plus_subset_non_zero_el_prod_maps_int.
rewrite plus_subset_non_zero_el_prod_maps_union.
unfold Symdiff_full. unfold Setminus_full.
unfold plus_subset_non_zero_el_prod_maps.
apply plus_set_functional. reflexivity.
Qed.



Lemma closed_all_plus_subsets_non_zero_el_prod_maps : 
  forall (E:Ensemble At) (pfe:Finite E),
    A_closed [a:At | exists (S:Ensemble (Fin_map E signe mns)) 
                            (pfi:Included S (non_zero_el_prod_maps E)),
                     a = plus_subset_non_zero_el_prod_maps E pfe S pfi].
intros E h1. 
constructor.
(* + *)
red. 
intros x y.
constructor. 
destruct x as [x h2].
destruct y as [y h3]. 
destruct h2 as [h2]. destruct h3 as [h3]. simpl.
destruct h2 as [X h2]. destruct h3 as [Y h3].
destruct h2 as [h2a h2b]. destruct h3 as [h3a h3b].
exists (Union X Y).
exists (inclusion_plus_subset_non_zero_el_prod_maps_preserves_union _ h1 _ _ h2a h3a).
unfold Bplus_sub. simpl.
rewrite h2b. rewrite h3b.
apply plus_subset_non_zero_el_prod_maps_union.
(* * *)
red. intros x y. constructor.
destruct x as [x h2].
destruct y as [y h3]. 
destruct h2 as [h2]. destruct h3 as [h3]. simpl.
destruct h2 as [X h2]. destruct h3 as [Y h3].
destruct h2 as [h2a h2b]. destruct h3 as [h3a h3b].
exists (Intersection X Y).
exists (inclusion_plus_subset_non_zero_el_prod_maps_preserves_int _ h1 _ _ h2a h3a).
unfold Btimes_sub. simpl.
rewrite h2b. rewrite h3b.
apply plus_subset_non_zero_el_prod_maps_int.
(* 1 *)
red. constructor.
exists (non_zero_el_prod_maps E). exists (inclusion_reflexive _).
symmetry.
apply plus_subset_non_zero_el_prod_maps_full.
(* 0 *)
red. constructor.
exists (Empty_set _). exists (empty_inclusion _).
symmetry.
apply plus_subset_non_zero_el_prod_maps_empty.
(* - *)
red. intro x. constructor.
destruct x as [x h2].
destruct h2 as [h2]. simpl.
destruct h2 as [X h2]. 
destruct h2 as [h2a h2b]. 
exists (Setminus (non_zero_el_prod_maps E) X).
exists (inclusion_plus_subset_non_zero_el_prod_maps_preserves_comp _ h1 X h2a).
unfold Bcomp_sub. simpl. 
rewrite h2b.
apply plus_subset_non_zero_el_prod_maps_comp.
Qed.

Lemma el_prod_sing_inc : 
  forall (E:Ensemble At) (pf:Finite E) (a:Fin_map E signe mns),
    el_prod a <> 0 -> Included (Singleton a) (non_zero_el_prod_maps E). 
intros E h1 a h2.
red.
intros a' h3.
destruct h3; subst.
unfold non_zero_el_prod_maps. constructor.
assumption.
Qed.

Lemma el_prod_sing : forall
  (E:Ensemble At) (pf:Finite E) (a:Fin_map E signe mns)
  (pfel:el_prod a <> 0), el_prod a = plus_subset_non_zero_el_prod_maps _ pf (Singleton a) (el_prod_sing_inc _ pf _ pfel).
intros E h1 a h2.
unfold plus_subset_non_zero_el_prod_maps.
assert (h3:Im (Singleton a) el_prod = (Singleton (el_prod a))).
  apply Extensionality_Ensembles.
  red. split.
  red. intros x' h13.
  inversion h13. subst. destruct H. subst. constructor.
  red. intros x' h13. destruct h13.
  apply Im_intro with a. constructor.
  reflexivity.
generalize dependent (
     (finite_image (Fin_map E signe mns) At (Singleton a) el_prod
        (Finite_downward_closed (Fin_map E signe mns)
           (non_zero_el_prod_maps E) (non_zero_el_prod_maps_fin E h1)
           (Singleton a) (el_prod_sing_inc E h1 a h2)))).
rewrite h3.
intro h4.
rewrite plus_set_sing'.
reflexivity.
Qed.


Lemma atom_non_zero_el_prod : 
  forall (E:Ensemble At) (pf:Finite E), 
    let C := (SubAlg _
(closed_all_plus_subsets_non_zero_el_prod_maps E pf))
  in forall (atm:Btype (Bc C)), atom atm <-> 
                              exists a:Fin_map E signe mns,
                                Ensembles.In (non_zero_el_prod_maps E) a /\ 
                                proj1_sig atm = el_prod a.
intros E h1 C atm.
split.
intros h2.
pose proof h2 as h3. 
rewrite atom_iff in h3.
pose proof (proj2_sig atm) as h4.
destruct h4 as [h4].
destruct h4 as [X h4]. destruct h4 as [h4 h5].
assert (h6: forall Y:Ensemble (Fin_map E signe mns), 
              Included Y (non_zero_el_prod_maps E) ->
                          Intersection Y X = X \/
                          Intersection Y X = Empty_set _).
  intros Y h7.
  assert (h8:Ensembles.In 
                [a : At
         | exists
             (S : Ensemble (Fin_map E signe mns)) (pfi : 
                                                  Included S
                                                  (non_zero_el_prod_maps E)),
             a = plus_subset_non_zero_el_prod_maps E h1 S pfi]
(plus_subset_non_zero_el_prod_maps E h1 _ h7)).
  constructor. exists Y, h7. reflexivity.
  pose (exist _ _ h8) as pY.
  red in h2.
  destruct h2 as [h2l h2r].
  specialize (h3 pY).
  destruct h3 as [h3l h3r].  
   destruct h3l as [h3a | h3b].
 (* h3a *)

  red in h3a. rewrite eq_ord in h3a. 
  simpl in h3a. unfold Btimes_sub in h3a. rewrite unfold_sig in h3a.
  pose proof (exist_injective _ _ _ _ _ h3a) as h6.
  clear h3a. unfold pY in h6. rewrite h5 in h6.
  simpl in h6.
  rewrite plus_subset_non_zero_el_prod_maps_int in h6.
  left.
  assert (h9:Included (Intersection X Y) X). auto with sets.
  rewrite comm_prod_psa.
  apply NNPP.
  intro h10.
  assert (h11:Strict_Included (Intersection X Y) X).  auto with sets.
  pose proof (Strict_Included_inv _ _ _ h11) as h12.
  destruct h12 as [h12l h12r]. clear h12l.
  rewrite setminus_same_int in h12r.  
  pose proof (Inhabited_not_empty _ _ h12r) as h13.
  pose proof (plus_subset_non_zero_el_prod_maps_empty_rev).

  assert (h14:Included (Setminus X Y) (non_zero_el_prod_maps E)).
    red. intros x h13'. destruct h13'. auto with sets.
 
  pose proof (plus_subset_non_zero_el_prod_maps_empty_rev _ h1 _ h14) as h15.
  assert (h16:plus_subset_non_zero_el_prod_maps E h1 (Setminus X Y) h14 <> 0). tauto.
(* maybe move a general version of this to SetUtilities.*)
  assert (h17:Setminus X Y = Intersection X (Setminus (non_zero_el_prod_maps E) Y)).
    apply Extensionality_Ensembles.
    red. split.
    red. 
    intros b h17.
    destruct h17 as [h17 h18].
    constructor; auto with sets.
    red. intros b h17.
    destruct h17 as [b h17 h18].
    constructor; auto with sets.
    destruct h18. assumption. 
  pose proof (inclusion_plus_subset_non_zero_el_prod_maps_preserves_comp _ h1 _ h7) as h18.
  assert (h19:plus_subset_non_zero_el_prod_maps E h1 (Setminus X Y) h14 = plus_subset_non_zero_el_prod_maps E h1 _ (inclusion_plus_subset_non_zero_el_prod_maps_preserves_int _ h1 _ _ h4 (inclusion_plus_subset_non_zero_el_prod_maps_preserves_comp _ h1 _ h7))).
  unfold plus_subset_non_zero_el_prod_maps.
  apply plus_set_functional. apply feq_im. assumption.
  rewrite h19 in h16.
  rewrite <- plus_subset_non_zero_el_prod_maps_int in h16.
  
  rewrite <- plus_subset_non_zero_el_prod_maps_comp in h16.
  unfold At in h16. rewrite <- le_iff in h16.
  unfold le in h16. rewrite eq_ord in h16.
  rewrite <- plus_subset_non_zero_el_prod_maps_int in h6.
  contradiction.
  (* h3b *)
  right.
  do 2 rewrite unfold_sig in h3b. unfold pY in h3b. simpl in h3b.
  pose proof (exist_injective _ _ _  _ _ h3b) as h6.
  unfold Btimes_sub in h6.
  simpl in h6.
  rewrite h5 in h6.
  rewrite plus_subset_non_zero_el_prod_maps_int in h6.
  rewrite comm_prod_psa.
  apply (plus_subset_non_zero_el_prod_maps_empty_rev _ h1 _ 
         (inclusion_plus_subset_non_zero_el_prod_maps_preserves_int _ h1 _ _ h4 h7)). assumption.
(* new goal *)
assert (h9:X <> Empty_set _).
  intro h10.
  generalize dependent h4.
  rewrite h10. intro h4.
  assert (h11:h4 = (empty_inclusion (non_zero_el_prod_maps E))).
    apply proof_irrelevance.
  rewrite h11.
  rewrite plus_subset_non_zero_el_prod_maps_empty.
  intro h12.
  red in h2. destruct h2 as [h2l h2r].  
  contradict h2l.
  rewrite (unfold_sig _ atm).
  rewrite (unfold_sig _ (Bzero (Bc C))).
  apply proj1_sig_injective.
  simpl. assumption.
pose proof (non_zero_el_prod_maps_fin _ h1) as h10.

pose proof (atom_sing _ _ h10 h4 h9) as h11.
rewrite h11 in h6.
destruct h6 as [a h6].
exists a.
destruct h6 as [h6l h6r].
generalize dependent h4.
rewrite h6r.
intros h4 h5. 
unfold plus_subset_non_zero_el_prod_maps in h5.
assert (h12:Im (Singleton a) (el_prod) = Singleton (el_prod a)).
  apply Extensionality_Ensembles.
  red. split.
  red. intros x' h13.
  inversion h13. subst. destruct H. subst. constructor.
  red. intros x' h13. destruct h13.
  apply Im_intro with a. constructor.
  reflexivity.
generalize dependent  (finite_image (Fin_map E signe mns) At (Singleton a) el_prod
            (Finite_downward_closed (Fin_map E signe mns)
               (non_zero_el_prod_maps E) (non_zero_el_prod_maps_fin E h1)
               (Singleton a) h4)).
rewrite h12. intros h13 h14.
rewrite plus_set_sing' in h14.
split; auto.
intro h2.
destruct h2 as [a h2]. 
pose proof h2 as h2'.
rewrite atom_iff.
intro b. split.
pose proof (proj2_sig b) as h3. simpl in h3.
destruct h3 as [h3].
destruct h3 as [S h3].
destruct h3 as [h3 h4].
pose proof (proj2_sig atm) as h5. simpl in h5.
destruct h5 as [h5]. destruct h5 as [A' h5]. destruct h5 as [h5 h6].
destruct (classic (A' = Empty_set _)) as [h7 | h8].
generalize dependent h5.
rewrite h7. intros h5 h6.
assert (h8:h5 = (empty_inclusion (non_zero_el_prod_maps E))).
  apply proof_irrelevance.
rewrite h8 in h6.
rewrite plus_subset_non_zero_el_prod_maps_empty in h6.
right.
rewrite (unfold_sig _ atm). rewrite (unfold_sig _ b).
rewrite (unfold_sig _ (Bzero (Bc C))).
simpl.
apply proj1_sig_injective. simpl. unfold Btimes_sub. simpl.
rewrite h6. rewrite comm_prod. apply zero_prod.
assert (h9:plus_subset_non_zero_el_prod_maps E h1 A' h5 <> 0).
  intro h10.
  pose proof plus_subset_non_zero_el_prod_maps_empty_rev.
  pose proof (plus_subset_non_zero_el_prod_maps_empty_rev _ h1 _ h5 h10) as h11.
contradiction.
rewrite <- h6 in h9.
destruct h2 as [h2'' h2].
rewrite h2 in h9.
pose proof el_prod_sing.
pose proof (el_prod_sing _ h1 a h9) as h10.
rewrite h10 in h2.
pose proof (@plus_subset_non_zero_el_prod_maps_int).
pose proof 
     (plus_subset_non_zero_el_prod_maps_int 
        _ h1 _ _ 
        (el_prod_sing_inc E h1 a h9) h3) as h11.
rewrite <- h2 in h11. rewrite <- h4 in h11.
pose proof (singleton_inc_int a S) as h12.
destruct h12 as [h12l | h12r].
generalize dependent (inclusion_plus_subset_non_zero_el_prod_maps_preserves_int E h1
             (Singleton a) S (el_prod_sing_inc E h1 a h9) h3).
rewrite comm_prod_psa in h12l.
rewrite h12l.
intros h13 h14.
assert (h15:h13 = (empty_inclusion (non_zero_el_prod_maps E))).
  apply proof_irrelevance.
rewrite h15 in h14.
rewrite plus_subset_non_zero_el_prod_maps_empty in h14.
right.
rewrite (unfold_sig _ atm). rewrite (unfold_sig _ b).  rewrite (unfold_sig _ (Bzero (Bc C))).
simpl.
apply proj1_sig_injective. simpl. unfold Btimes_sub. simpl.
assumption. 
left.
red. rewrite eq_ord.
generalize dependent (inclusion_plus_subset_non_zero_el_prod_maps_preserves_int E h1
             (Singleton a) S (el_prod_sing_inc E h1 a h9) h3).
rewrite comm_prod_psa in h12r.
rewrite h12r. intros h13 h14.
assert (h15:h13 =  (el_prod_sing_inc E h1 a h9)). apply proof_irrelevance.
rewrite h15 in h14.
rewrite <- h2 in h14.
rewrite (unfold_sig _ atm). rewrite (unfold_sig _ b). simpl.
unfold Btimes_sub. simpl. apply proj1_sig_injective. simpl.
assumption.
(* new goal*)  
unfold le. rewrite eq_ord.
intro h3.
pose proof (proj2_sig b) as h4.
destruct h3 as [h3l h3r]. pose proof h3l as h3l'.
rewrite h3r in h3l.
destruct h2 as [h2l h2r].
rewrite (unfold_sig _ atm) in h3l.
rewrite (unfold_sig _ (Bzero (Bc C))) in h3l.
pose proof (exist_injective _ _ _ _ _ h3l) as h8.
simpl in h8.
rewrite <- h8 in h2r.
unfold non_zero_el_prod_maps in h2l.
destruct h2l as [h2l].
symmetry in h2r.
contradiction.
Qed.
  
Lemma plus_subset_non_zero_el_prod_maps_inj :
  forall (E:Ensemble At) (pfe:Finite E) 
         (S T:Ensemble (Fin_map E signe mns))
    (pfs:Included S (non_zero_el_prod_maps E))
    (pft:Included T (non_zero_el_prod_maps E)),
    plus_subset_non_zero_el_prod_maps E pfe S pfs =
    plus_subset_non_zero_el_prod_maps E pfe T pft ->
    S = T.
intros E h1 S T h2 h3 h4.
pose proof (plus_subset_non_zero_el_prod_maps_empty _ h1) as h5.
pose proof (symdiff_full_ref (non_zero_el_prod_maps E) S) as h6.
generalize dependent  (empty_inclusion (non_zero_el_prod_maps E)).
rewrite <- h6.
pose proof (plus_subset_non_zero_el_prod_maps_symdiff_full _ h1 _ _ h2 h2) as h7.
rewrite h4 in h7 at 2.
generalize dependent (inclusion_plus_subset_non_zero_el_prod_maps_preserves_symdiff_full
            E h1 S S h2 h2).
intros h7 h8 h9 h10.
assert (h11: h7 = h9).  apply proof_irrelevance. subst.
rewrite h10 in h8.
rewrite plus_subset_non_zero_el_prod_maps_symdiff_full in h8.
pose proof (inclusion_plus_subset_non_zero_el_prod_maps_preserves_symdiff_full E h1 S T h2 h3) as h11.
pose proof (plus_subset_non_zero_el_prod_maps_empty_rev _ h1 (Symdiff_full (non_zero_el_prod_maps E) S T) h11) as h12.
generalize dependent  (inclusion_plus_subset_non_zero_el_prod_maps_preserves_symdiff_full
            E h1 S T h2 h3). intro h13.
assert (h14:h11 = h13). apply proof_irrelevance. subst.
intro h14. specialize (h12 h14).
rewrite symdiff_full_empty_iff_eq in h12; auto.
Qed.

Lemma el_prod_inc_c : 
  forall (E F:Ensemble At) (pfe:Finite E)
         (b:Fin_map F signe mns), Included F E -> 
    let C := 
        [a:At | exists (S:Ensemble (Fin_map E signe mns)) 
                                           (pfi:Included S (non_zero_el_prod_maps E)),
                     a = plus_subset_non_zero_el_prod_maps E pfe S pfi] in
    Ensembles.In C (el_prod b).
intros E F h1 b h2 C.
constructor.
pose  [a : Fin_map E signe mns
      | Ensembles.In (non_zero_el_prod_maps E) a /\ extends a b] as L.
exists L.
assert (h3: Included L (non_zero_el_prod_maps E)).
  red. intros x h3. destruct h3 as [h3]. destruct h3; assumption.
exists h3. 
assert (h4:forall a:Fin_map E signe mns, 
             Ensembles.In L a -> 
             Included (Im F (fun i : At => eps i (b |-> i)))
                      (Im E (fun i : At => eps i (a |-> i)))).
  intros a h5. red. intros p h6.
  destruct h5 as [h5]. destruct h5 as [h5l h5r].
  red in h5r. destruct h5r as [h5a h5b].
  destruct h6 as [f h6 p]. subst.
  rewrite h5b; auto.
  apply Im_intro with f. auto with sets.
  reflexivity.
assert (h5:forall a:Fin_map E signe mns, 
             Ensembles.In L a -> le (el_prod a) (el_prod b)). 
  intros a h6.
 unfold el_prod.
  apply times_set_inc_le. apply h4; auto.
assert (h6:forall a:Fin_map E signe mns,
             Ensembles.In (Setminus (non_zero_el_prod_maps E) L) a ->
             (el_prod a) * (el_prod b) = 0).
  intros a h7.
  destruct h7 as [h7 h8].
  unfold L in h8.
  assert (h9:~extends a b).
    intro h10.
    assert (h11:Ensembles.In  [a0 : Fin_map E signe mns
         | Ensembles.In (non_zero_el_prod_maps E) a0 /\ extends a0 b] a).
      constructor. split; assumption.
      contradiction.
  pose proof (not_extends_differs_at_point _ _ h2 h9) as h10.
  destruct h10 as [i h10].
  destruct h10 as [h10l h10r].
  pose proof el_prod_le_ai.
  pose proof (el_prod_le_ai _ b _ h10l) as h11.
  assert (h12:Ensembles.In E i). auto with sets.
  pose proof (el_prod_le_ai _ a _ h12) as h13.
  pose proof (mono_prod _ _ _ _ h11 h13) as h14.
  destruct (b |-> i); destruct (a |-> i). 
  contradict h10r. reflexivity.
  simpl in h14. rewrite comp_prod in h14. apply le_x_0. 
  rewrite comm_prod. assumption.
  simpl in h14. rewrite (comm_prod _ (-i) i) in h14. 
  rewrite comp_prod in h14. apply le_x_0.  
  rewrite comm_prod. assumption. contradict h10r. reflexivity.
assert (h7:el_prod b = (el_prod b) * 1). rewrite one_prod. reflexivity.
rewrite <- (plus_subset_non_zero_el_prod_maps_full _ h1) in h7.
pose proof (decompose_setminus_inc _ _ h3) as h8.
assert (h9: plus_subset_non_zero_el_prod_maps E h1 (non_zero_el_prod_maps E)
         (inclusion_reflexive (non_zero_el_prod_maps E)) = 
            plus_subset_non_zero_el_prod_maps E h1 (Union L (Setminus (non_zero_el_prod_maps E) L)) (inclusion_plus_subset_non_zero_el_prod_maps_preserves_union _ h1 _ _ h3 (inclusion_plus_subset_non_zero_el_prod_maps_preserves_comp _ h1 _ h3))).
  unfold plus_subset_non_zero_el_prod_maps.
  apply plus_set_functional.
  apply feq_im. assumption.
rewrite h9 in h7.
unfold plus_subset_non_zero_el_prod_maps in h7. 
rewrite dist_set_plus1 in h7.
generalize  dependent (finite_image (Btype (Bc A)) (Btype (Bc A))
            (Im (Union L (Setminus (non_zero_el_prod_maps E) L)) el_prod)
            (fun y : Btype (Bc A) => el_prod b * y)
            (finite_image (Fin_map E signe mns) At
               (Union L (Setminus (non_zero_el_prod_maps E) L)) el_prod
               (Finite_downward_closed (Fin_map E signe mns)
                  (non_zero_el_prod_maps E) (non_zero_el_prod_maps_fin E h1)
                  (Union L (Setminus (non_zero_el_prod_maps E) L))
                  (inclusion_plus_subset_non_zero_el_prod_maps_preserves_union
                     E h1 L (Setminus (non_zero_el_prod_maps E) L) h3
                     (inclusion_plus_subset_non_zero_el_prod_maps_preserves_comp
                        E h1 L h3))))).

rewrite im_im.  
rewrite image_union. 
intros h10 h10'.
assert (h11:Included (Im L (fun x:Fin_map E signe mns => el_prod b * el_prod x))  (Union (Im L  (fun x:Fin_map E signe mns => el_prod b * el_prod x)) (Im (Setminus (non_zero_el_prod_maps E) L)  (fun x:Fin_map E signe mns => el_prod b * el_prod x)))).
  auto with sets.
assert (h12:Included  (Im (Setminus (non_zero_el_prod_maps E) L)  (fun x:Fin_map E signe mns => el_prod b * el_prod x))  (Union (Im L  (fun x:Fin_map E signe mns => el_prod b * el_prod x))
             (Im (Setminus (non_zero_el_prod_maps E) L)  (fun x:Fin_map E signe mns => el_prod b * el_prod x)))).
  auto with sets.
assert (h13:Finite (Im L  (fun x:Fin_map E signe mns => el_prod b * el_prod x))).  eapply Finite_downward_closed. apply h10. apply h11.
assert (h14:Finite (Im (Setminus (non_zero_el_prod_maps E) L)  (fun x:Fin_map E signe mns => el_prod b * el_prod x))). eapply Finite_downward_closed. apply h10. apply h12.  
rewrite (plus_set_union' _ _ _ h13 h14) in h10'.
assert (h16:plus_set (Im (Setminus (non_zero_el_prod_maps E) L)  (fun x:Fin_map E signe mns => el_prod b * el_prod x)) h14 = 0).
  destruct (classic (Inhabited (Setminus (non_zero_el_prod_maps E) L))) as [h17 | h18].
    rewrite <- (plus_set_sing _ 0).
    apply plus_set_functional.
    apply Extensionality_Ensembles.
    red. split.
    red.
    intros x h18.
    destruct h18 as [x h18]. subst.
    rewrite comm_prod.
    rewrite h6; auto. constructor.
    red.
    intros z h18.
    destruct h18; subst.
    destruct h17 as [a h17].
    apply Im_intro with a; auto.
    rewrite comm_prod. rewrite h6; auto.
    pose proof (not_inhabited_empty _ h18) as h19.
    pose proof (image_empty _ _ (fun x : Fin_map E signe mns => el_prod b * el_prod x)) as h20.
    rewrite <- h19 in h20.
    rewrite <- plus_set_empty.
    apply plus_set_functional.
    assumption.
rewrite h10'.
rewrite h16.
rewrite zero_sum.
unfold plus_subset_non_zero_el_prod_maps.
apply plus_set_functional.
pose proof h5 as h5'. clear h5.
apply Extensionality_Ensembles. red.
split.
red.
intros x h17.
destruct h17 as [x h17]. subst.

unfold le in h5'. specialize (h5' x h17). rewrite eq_ord in h5'.
rewrite comm_prod.
rewrite h5'. apply Im_intro with x. assumption. reflexivity.
red. 
intros x h17. destruct h17 as [x h17]. subst.
specialize (h5' x h17). unfold le in h5'. rewrite eq_ord in h5'.
rewrite <- h5'.
rewrite comm_prod.
apply Im_intro with x. assumption. reflexivity.
Qed.


Lemma el_prod_empty : forall (F:Fin_map (Empty_set At) signe mns),
                        el_prod F = 1.
intros F. unfold el_prod.
generalize dependent (finite_image At At (Empty_set At) (fun i : At => eps i (F |-> i))
        (Fin_map_fin_dom F)).
rewrite image_empty.
intro h1.
apply times_set_empty'.
Qed.


Lemma el_prod_decompose_restriction : 
  forall (E F:Ensemble At) (pf:Included F E)
         (a:Fin_map E signe mns),
  el_prod a = el_prod (restriction a pf) * 
              el_prod (restriction a (setminus_inc E F)).
intros E F h1 a.
unfold el_prod.
pose proof (decompose_setminus_inc _ _ h1) as h2.
generalize  (finite_image At At E (fun i : At => eps i (a |-> i))
        (Fin_map_fin_dom a)).
pose proof (feq_im _ _ (fun i : At => eps i (a |-> i)) h2) as h3.
rewrite image_union in h3.
rewrite h3.
intro h4. 
rewrite <- times_set_union.
apply times_set_functional.
apply Extensionality_Ensembles.
red. split.
red.
intros y h5.
destruct h5 as [y h5l | y h5r].
left. 
destruct h5l as [x h5l y h6].  rewrite h6.
apply Im_intro with x. assumption.
f_equal.
rewrite restriction_compat; auto.
right.
destruct h5r as [x h5l y h6].  rewrite h6.
apply Im_intro with x. assumption.
f_equal. rewrite restriction_compat; auto.
red.
intros y h5.
destruct h5 as [y h5l | y h5r].
left.
destruct h5l as [x h5l y h6]. rewrite h6.
apply Im_intro with x; auto. f_equal. rewrite restriction_compat; auto.
destruct h5r as [x h5r y h6]. rewrite h6. right.
apply Im_intro with x; auto. f_equal. rewrite restriction_compat; auto.
Qed.

Lemma eps_in_closed :
  forall (S:Ensemble At),
    A_closed S ->
    forall x:At,
      Ensembles.In S x ->
      forall s:sign, Ensembles.In S (eps x s).
intros S h1 x h2 s.
destruct s. simpl. assumption. simpl.
pose proof (C_c _ _ h1) as h3.
red in h3.
specialize (h3 (exist _ _ h2)).
unfold Bcomp_sub in h3.
simpl in h3.
assumption.
Qed.

Lemma el_prod_in_gen : 
  forall (E:Ensemble At)
         (F:Fin_map E signe mns),
    Ensembles.In (Gen_Ens E) (el_prod F).
intros E F.
pose proof (Fin_map_fin_dom F) as h1.
revert F.
induction h1 as [|E h1 h2 x h3].
intros F.
rewrite gen_ens_empty.
rewrite el_prod_empty. right.
intro F.
assert (h4:Included E (Add E x)). auto with sets.
specialize (h2 (restriction F h4)).
pose proof (el_prod_decompose_restriction).
rewrite (el_prod_decompose_restriction _ _ h4).
assert (h5: el_prod (restriction F (setminus_inc (Add E x) E)) =
            eps x (F |-> x)).
  unfold el_prod.
  generalize  (finite_image At At (Setminus (Add E x) E)
        (fun i : At => eps i (restriction F (setminus_inc (Add E x) E) |-> i))
        (Fin_map_fin_dom (restriction F (setminus_inc (Add E x) E)))).

  pose proof (setminus_add1 _ _ h3) as h5.
pose proof (feq_im _ _ (fun i : At => eps i (restriction F (setminus_inc (Add E x) E) |-> i)) h5) as h6.
rewrite h6.
rewrite im_singleton.
intro h7.
rewrite times_set_sing'.
f_equal.
rewrite restriction_compat. reflexivity. constructor. right.
constructor. assumption.
rewrite h5.
pose proof (gen_ens_preserves_inclusion _ _ _ h4) as h6.
assert (h7:Ensembles.In (Gen_Ens (Add E x)) (el_prod (restriction F h4))). auto with sets.
pose proof (closed_gen_ens _ (Add E x)) as h8.
assert (h9:Ensembles.In (Gen_Ens (Add E x)) x).
  constructor. intros S h9.
  destruct h9 as [h9]. destruct h9 as [h9l h9r].
  pose proof (Add_intro2 _ E x). auto with sets.
pose proof (eps_in_closed _ h8 _ h9 (F |-> x)) as h10.
apply times_closed; auto.
Qed.

Lemma plus_subset_non_zero_el_prod_maps_in_gen : 
  forall (E:Ensemble At) (pf:Finite E) 
         (S:Ensemble (Fin_map E signe mns))
         (pfi:Included S (non_zero_el_prod_maps E)),
    Ensembles.In (Gen_Ens E) (plus_subset_non_zero_el_prod_maps E pf S pfi).
intros E h1 S.
pose proof (finite_fin_map_ens S h1 signe_finite) as h2.
induction h2 as [|S h3 h4 a h5].
intro h2.
unfold plus_subset_non_zero_el_prod_maps.
generalize  (finite_image (Fin_map E signe mns) At
           (Empty_set (Fin_map E signe mns)) el_prod
           (Finite_downward_closed (Fin_map E signe mns)
              (non_zero_el_prod_maps E) (non_zero_el_prod_maps_fin E h1)
              (Empty_set (Fin_map E signe mns)) h2)).
rewrite image_empty.
intro h3. rewrite plus_set_empty'.
pose proof (closed_gen_ens _ E) as h4.
apply zero_closed; auto.
intro h6.
assert (h7:Included S (Add S a)). auto with sets.
assert (h8:Included S (non_zero_el_prod_maps E)). auto with sets.
specialize (h4 h8).
unfold plus_subset_non_zero_el_prod_maps.
unfold Add.
generalize (finite_image (Fin_map E signe mns) At (Union S (Singleton a))
           el_prod
           (Finite_downward_closed (Fin_map E signe mns)
              (non_zero_el_prod_maps E) (non_zero_el_prod_maps_fin E h1)
              (Union S (Singleton a)) h6)).
rewrite image_union.
intro h9. 
unfold plus_subset_non_zero_el_prod_maps in h4.
pose proof (finite_image _ _ _ el_prod h3) as h10.
pose proof (Singleton_is_finite _ a) as h11.
pose proof (finite_image _ _ _ el_prod h11) as h12.
pose proof (plus_set_union' _ _ _ h10 h12 h9) as h13.
unfold At in h13. unfold At.
rewrite h13.
assert (h14:(finite_image (Fin_map E signe mns) At S el_prod
               (Finite_downward_closed (Fin_map E signe mns)
                  (non_zero_el_prod_maps E) (non_zero_el_prod_maps_fin E h1)
                  S h8)) = h10). apply proof_irrelevance.
rewrite h14 in h4.
pose proof (closed_gen_ens _ E) as h15.
apply plus_closed; auto. 
assert (h16: (plus_set (Im (Singleton a) el_prod) h12) = el_prod a).
  rewrite <- plus_set_sing.
  apply plus_set_functional.
  rewrite im_singleton. reflexivity.
unfold At in h16.
rewrite h16.
apply el_prod_in_gen.
Qed.


Lemma gen_e_eq_c : 
  forall (E:Ensemble At) (pf:Finite E), 
    let C :=
        [a:At | exists (S:Ensemble (Fin_map E signe mns))
                       (pfi:Included S (non_zero_el_prod_maps E)),
                a = plus_subset_non_zero_el_prod_maps E pf S pfi] in
    Gen_Ens E = C.
intros E h1 C.
apply Extensionality_Ensembles.
red. split.
(* <= *)
assert (h2:Included E C).
  red.
  intros x h2.
  pose proof (In_singleton _ x) as h3.
  assert (h4:Included (Singleton x) E). red. intros x' h4.
    destruct h4; subst. assumption.
  assert (h6:functionally_paired (Singleton x) signe (Singleton (x, pls))).
    constructor.
    intros x' h5.
    exists pls. red. split. split.
    unfold signe. left.
    destruct h5; subst. constructor.
    intros s h6.
    destruct h6 as [h6l h6r].
    inversion h6r. reflexivity.
    intros pr h5. 
    inversion h5 as [h6]. rewrite (surjective_pairing pr) in h6.
    inversion h6. subst.
    simpl. split. constructor. left.
  pose proof (Singleton_is_finite _ x) as h7.
  pose (fin_map_intro _ _ mns h7 signe_finite _ h6) as F.
  assert (h8: x = el_prod F).
    unfold F. unfold el_prod.
    generalize (finite_image At At (Singleton x)
        (fun i : At =>
         eps i
           (fin_map_intro (Singleton x) signe mns h7 signe_finite
              (Singleton (x, pls)) h6 |-> i))
        (Fin_map_fin_dom
           (fin_map_intro (Singleton x) signe mns h7 signe_finite
              (Singleton (x, pls)) h6))).
   rewrite im_singleton. intro h8.
   rewrite times_set_sing'.
   simpl.
   pose proof @fps_to_f_s_compat.
   pose proof (fps_to_f_s_compat h6 mns _ h3) as h9.
   inversion h9. simpl. reflexivity.
  rewrite h8.
  apply el_prod_inc_c. assumption.
pose proof gen_ens_preserves_inclusion.
pose proof (gen_ens_preserves_inclusion _ _ _ h2) as h3.
rewrite <- gen_ens_closed_eq.
assumption.
apply closed_all_plus_subsets_non_zero_el_prod_maps.
(* >=*)
red.
intros x h2.
destruct h2 as [h2].
destruct h2 as [S h2].
destruct h2 as [h2 h3].
subst.
apply plus_subset_non_zero_el_prod_maps_in_gen.
Qed.



(* The main theorem of this section, along with unprimed eponym!*)
Theorem normal_form' :
  forall (E:Ensemble At) (pfe:Finite E),
    let Bt := (Btype (Bc (Gen E))) in
    (forall p:Bt, atom p <-> 
                      ((p <> 0) /\ 
                       exists f:Fin_map E signe mns,
                         proj1_sig p = el_prod f))
                         /\
    (forall b:Bt, exists !(S:Ensemble (Fin_map E signe mns)),
       exists (pf:Included S (non_zero_el_prod_maps E)),
       plus_subset_non_zero_el_prod_maps E pfe S pf =
       proj1_sig b).
intros E h1 Bt.
pose proof (gen_e_eq_c _ h1) as h2.
pose proof (closed_gen_ens _ E) as h3.
pose proof (closed_all_plus_subsets_non_zero_el_prod_maps _ h1) as h4.
unfold Bt. unfold Gen.  
pose (fun (D:Ensemble At) (pf:A_closed D) =>
 (forall p : Btype (Bc (SubAlg D pf)),
    atom p <->
    p <> 0 /\ (exists f : Fin_map E signe mns, proj1_sig p = el_prod f)) /\
   (forall b : Btype (Bc (SubAlg D pf)),
    exists ! S : Ensemble (Fin_map E signe mns),
      exists pf : Included S (non_zero_el_prod_maps E),
        plus_subset_non_zero_el_prod_maps E h1 S pf = proj1_sig b)) as f.
assert (h5:f _ (closed_gen_ens _ E) <-> f _ h4).  
  generalize (closed_gen_ens _ E). rewrite h2.  intro h6. 
  assert (h7:h4 = h6). apply proof_irrelevance. subst.
  tauto.
unfold f in h5. simpl in h5. simpl. 
rewrite h5.
split.
intro p. split.
intro h8. 
pose proof (atom_non_zero_el_prod _ h1 p) as h9.
unfold atom in h8. unfold atom in h9.
assert (h4 =  (closed_all_plus_subsets_non_zero_el_prod_maps E h1)). apply proof_irrelevance. subst. 
pose proof h8 as h8'. destruct h8' as [h8l h8r].
simpl in h8l. split. unfold At in h8l. unfold At.
assumption.
rewrite h9 in h8. 
destruct h8 as [a h8].
exists a. destruct h8; assumption.
intro h6.
pose proof (atom_non_zero_el_prod _ h1 p) as h9.
destruct h6 as [h6l h6r].
assert (h10: (exists a : Fin_map E signe mns,
          Ensembles.In (non_zero_el_prod_maps E) a /\ proj1_sig p = el_prod a)).
  destruct h6r as [a h6r].
  exists a. split.
  constructor.
  rewrite <- h6r.
  intro h10. 
  rewrite (unfold_sig _ p) in h6l.
  contradict h6l. apply proj1_sig_injective. simpl. assumption.
  assumption.
rewrite <- h9 in h10.
unfold atom. unfold atom in h10.
assert (h4 = (closed_all_plus_subsets_non_zero_el_prod_maps E h1)). apply proof_irrelevance. subst.
assumption. 
intro b.
pose proof (proj2_sig b) as h7.
simpl in h7.
destruct h7 as [h7].
destruct h7 as [S [h7 h8]].
exists S. red.
split.
exists h7. symmetry. assumption.
intros S' h10.
destruct h10 as [h10l h10r].
rewrite h8 in h10r.
symmetry in h10r.
apply (plus_subset_non_zero_el_prod_maps_inj _ h1 _ _ h7 h10l h10r).
Qed.


Lemma plus_set_inc_sig : 
  forall (E F:Ensemble At) (pfi:Included E F) (pfe:Finite E) 
         (pfss:Finite (inc_sig E F pfi)) (pff:A_closed F),
         proj1_sig (@plus_set (SubAlg _ pff) (inc_sig E _ pfi) pfss) =
         plus_set E pfe. 
intros E F h1 h2 h3 h4. 
assert (h5:E = Im (inc_sig E F h1) (@proj1_sig _ _)).
  apply Extensionality_Ensembles.
  red. split.
  red.
  intros x h5.
  apply Im_intro with (exist _ _ (h1 _ h5)).
  constructor. simpl. reflexivity.
  red. intros x h5.
  destruct h5 as [x h5]. subst.
  rewrite <- (inc_sig_iff _ _ h1). assumption.  
pose proof h2 as h2'.
rewrite h5 in h2'. 
assert (h6:plus_set E h2 = plus_set _ h2').
  apply plus_set_functional. assumption.
rewrite h6.
unfold At.
pose proof h2 as h2''.
rewrite <- (finite_inc_sig_iff _ _ h1) in h2''. 
pose proof (plus_set_sub_compat _ _ h4 _ h2'' h2') as h9.
assert (h10:h2'' = h3). apply proof_irrelevance. 
rewrite h10 in h9.
unfold At. unfold At in h9.
rewrite <- h9.
apply plus_set_functional.
reflexivity.
Qed.

Lemma times_set_inc_sig : 
  forall (E F:Ensemble At) (pfi:Included E F) (pfe:Finite E) 
         (pfss:Finite (inc_sig E F pfi)) (pff:A_closed F),
         proj1_sig (@times_set (SubAlg _ pff) (inc_sig E _ pfi) pfss) =
         times_set E pfe. 
intros E F h1 h2 h3 h4. 
assert (h5:E = Im (inc_sig E F h1) (@proj1_sig _ _)).
  apply Extensionality_Ensembles.
  red. split.
  red.
  intros x h5.
  apply Im_intro with (exist _ _ (h1 _ h5)).
  constructor. simpl. reflexivity.
  red. intros x h5.
  destruct h5 as [x h5]. subst.
  rewrite <- (inc_sig_iff _ _ h1). assumption.  
pose proof h2 as h2'.
rewrite h5 in h2'. 
assert (h6:times_set E h2 = times_set _ h2').
  apply times_set_functional. assumption.
rewrite h6.
unfold At.
pose proof h2 as h2''.
rewrite <- (finite_inc_sig_iff _ _ h1) in h2''. 
pose proof (times_set_sub_compat _ _ h4 _ h2'' h2') as h9.
assert (h10:h2'' = h3). apply proof_irrelevance. 
rewrite h10 in h9.
unfold At. unfold At in h9.
rewrite <- h9.
apply times_set_functional.
reflexivity.
Qed.



(* More faithful to Givant/Halmos' form of the normal_form *)
Theorem normal_form :
  forall (E:Ensemble At),
    let Bt := (Btype (Bc (Gen E))) in
    Finite E ->
    (forall p:Bt, atom p <-> 
                      ((p <> 0) /\ 
                       exists f:Fin_map E signe mns,
                         proj1_sig p = el_prod f))
                         /\
    (forall b:Bt, exists !(S:Ensemble Bt),
       exists pf:Finite S, 
         (forall x:Bt, Ensembles.In S x -> atom x) /\
         plus_set S pf = b).
intros E Bt h1.
pose proof (normal_form' E h1) as h2.
destruct h2 as [h2l h2r].
split. assumption. 
intros b.
specialize (h2r b).
destruct h2r as [S h2r].
pose (Im S el_prod) as imS.
assert (h3':Included (Im S el_prod) (Gen_Ens E)).
  red.
  intros y h4.
  destruct h4 as [y h4 x]. subst.
  apply el_prod_in_gen.
exists (inc_sig _ _ h3').
red in h2r. red.
destruct h2r as [h2a h2b].
destruct h2a as [h6 h7].
pose proof (finite_fin_map_ens S h1 signe_finite) as h3.
pose proof (finite_image _ _ _ el_prod h3) as h4.  
rewrite <- (finite_inc_sig_iff _ _ h3') in h4.
split. 
exists h4.
split. intros x h5.
inversion h5 as [c h8].  subst.
destruct h8 as [a h8 y]. subst.
assert (h9: (exist (Ensembles.In (Gen_Ens E)) (el_prod a)
        (h3' (el_prod a)
           (Im_intro (Fin_map E signe mns) At S el_prod a h8 
              (el_prod a) eq_refl))) <> (Bzero (Bc (Gen E))) /\
            (exists f:Fin_map E signe mns, proj1_sig  (exist (Ensembles.In (Gen_Ens E)) (el_prod a)
        (h3' (el_prod a)
           (Im_intro (Fin_map E signe mns) At S el_prod a h8 
              (el_prod a) eq_refl))) = el_prod f)).
  split.
  assert (h10:Ensembles.In (non_zero_el_prod_maps E) a).  auto with sets.
  destruct h10 as [h10].
  intro h11.
  pose proof (exist_injective _ _ _ _ _ h11).
  contradiction.
  exists a. simpl. reflexivity.
rewrite <- h2l in h9.
assumption.
apply proj1_sig_injective.
rewrite <- h7.
unfold plus_subset_non_zero_el_prod_maps.
pose proof (finite_image _ _ _ el_prod h3) as h8.
pose proof (closed_gen_ens _ E) as h9.
pose proof (plus_set_inc_sig _ _ h3' h8 h4 h9) as h10.
assert (h11:h8 = (finite_image (Fin_map E signe mns) At S el_prod
        (Finite_downward_closed (Fin_map E signe mns)
           (non_zero_el_prod_maps E) (non_zero_el_prod_maps_fin E h1) S h6))). apply proof_irrelevance.
subst. 
rewrite <- h10. 
f_equal.
unfold Gen.
assert (h11:h9 = (closed_gen_ens A E)). apply proof_irrelevance.
subst.
apply plus_set_functional.
reflexivity.
intros D h6'.
destruct h6' as [h6l h6r].
destruct h6r as [h6a h6b].
specialize (h2b [a:(Fin_map E signe mns) | Ensembles.In D (exist _ (el_prod a) (el_prod_in_gen E a)) /\ el_prod a <> 0]). 
assert (h7': Included
                  [a : Fin_map E signe mns | Ensembles.In D (exist _ (el_prod a) (el_prod_in_gen E a)) /\ el_prod a <> 0]
                  (non_zero_el_prod_maps E)).
  red. intros a h7'.
  destruct h7' as [h7']. destruct h7'.
  constructor. assumption.
assert (h8:plus_subset_non_zero_el_prod_maps E h1
             [a : Fin_map E signe mns
             | Ensembles.In D (exist _ (el_prod a) (el_prod_in_gen E a)) /\ el_prod a <> 0] h7' = 
           proj1_sig b).
unfold plus_subset_non_zero_el_prod_maps.
assert (h8:Finite [a:Fin_map E signe mns | Ensembles.In D (exist _ (el_prod a) (el_prod_in_gen E a)) /\ el_prod a = 0]).
  apply finite_fin_map_ens. assumption. apply signe_finite.
pose proof (finite_image _ _ _ el_prod h8) as h9.
assert (h10:plus_set (Im [a:Fin_map E signe mns | Ensembles.In D (exist _ (el_prod a) (el_prod_in_gen E a)) /\ el_prod a = 0] el_prod) h9 = 0).
  destruct (classic (Inhabited [a:Fin_map E signe mns | Ensembles.In D (exist _ (el_prod a) (el_prod_in_gen E a)) /\ el_prod a = 0])) as [h10 | h11].
  rewrite <- plus_set_sing.
  apply plus_set_functional.
  apply Extensionality_Ensembles.
  red. split.
  red.
  intros x h11. destruct h11 as [x h11 x' h12]. subst.
  destruct h11 as [h11]. destruct h11 as [h11a h11b].
  rewrite h11b. constructor.
  red. intros x h11. destruct h11; subst.
  destruct h10 as [x h10].
  destruct h10 as [h10].
  apply Im_intro with x. constructor. assumption. symmetry. destruct h10; auto.
  pose proof (not_inhabited_empty  _ h11) as h12.
  rewrite <- plus_set_empty.
  apply plus_set_functional.
  pose proof (feq_im _ _ el_prod h12) as h13.
  rewrite image_empty  in h13.
  assumption.
pose proof (f_equal (Bplus _  (plus_set
     (Im
        [a : Fin_map E signe mns
        | Ensembles.In D
            (exist (Ensembles.In (Gen_Ens E)) (el_prod a)
               (el_prod_in_gen E a)) /\ el_prod a <> 0] el_prod)
     (finite_image (Fin_map E signe mns) At
        [a : Fin_map E signe mns
        | Ensembles.In D
            (exist (Ensembles.In (Gen_Ens E)) (el_prod a)
               (el_prod_in_gen E a)) /\ el_prod a <> 0] el_prod
        (Finite_downward_closed (Fin_map E signe mns)
           (non_zero_el_prod_maps E) (non_zero_el_prod_maps_fin E h1)
           [a : Fin_map E signe mns
           | Ensembles.In D
               (exist (Ensembles.In (Gen_Ens E)) (el_prod a)
                  (el_prod_in_gen E a)) /\ el_prod a <> 0] h7')))) h10) as h11.
rewrite zero_sum in h11. 
unfold At in h11. unfold At.
rewrite <- h11.
rewrite <- plus_set_union.
rewrite <- h6b.
pose proof (finite_image _ _ _ (@proj1_sig _ _) h6l) as h12.
pose proof (plus_set_sub_compat _ _ (closed_gen_ens _ E) _ h6l h12) as h13.
unfold At in h13. unfold At.
unfold Gen.
rewrite <- h13. unfold Gen in h11.
apply plus_set_functional.

assert (h12':(Union
        [a : Fin_map E signe mns
        | Ensembles.In D (exist _ (el_prod a) (el_prod_in_gen E a)) /\ el_prod a <> 0]
        [a : Fin_map E signe mns
        | Ensembles.In D (exist _ (el_prod a) (el_prod_in_gen E a)) /\ el_prod a = 0]) =
            [a : Fin_map E signe mns | Ensembles.In D (exist _ (el_prod a) (el_prod_in_gen E a))]).
  apply Extensionality_Ensembles.
  red. split. red. 
  intros x h12'.
  destruct h12' as [x h12l | x h12r].
  destruct h12l as [h12l]. destruct h12l; constructor; auto.
  destruct h12r as [h12r]. destruct h12r; constructor; auto.
  red. intros x h12'.
  destruct h12' as [h12'].
  destruct (classic_dec (el_prod x = 0)) as [h13' | h14'].
  right. constructor; auto. left. constructor; auto.
rewrite <- image_union.
unfold At in h12'. unfold At.  

rewrite h12'. 
  apply Extensionality_Ensembles.
  red. split. red.
  intros x h13'.
  destruct h13' as [x h13']. subst.
  destruct h13' as [h13']. 
  apply Im_intro with (exist (Ensembles.In (Gen_Ens E)) (el_prod x) (el_prod_in_gen E x)).
assumption.  simpl. reflexivity.
red.
intros x h14.
destruct h14 as [x h14]. subst.
pose proof h6a as h15. clear h6a. 
specialize (h15 _ h14).
rewrite h2l in h15.
destruct h15 as [h15l h15r].
destruct h15r as [a h15r].
apply Im_intro with a.
constructor. 
assert (h16:(exist (Ensembles.In (Gen_Ens E)) (el_prod a) (el_prod_in_gen E a)) = x).
apply proj1_sig_injective. simpl.
rewrite <- h15r. reflexivity.
rewrite h16.
assumption.
assumption.
assert (h9:(exists
           pf : Included
                  [a : Fin_map E signe mns
                  | Ensembles.In D
                      (exist (Ensembles.In (Gen_Ens E)) 
                         (el_prod a) (el_prod_in_gen E a)) /\ 
                    el_prod a <> 0] (non_zero_el_prod_maps E),
           plus_subset_non_zero_el_prod_maps E h1
             [a : Fin_map E signe mns
             | Ensembles.In D
                 (exist (Ensembles.In (Gen_Ens E)) 
                    (el_prod a) (el_prod_in_gen E a)) /\ 
               el_prod a <> 0] pf = proj1_sig b)).
  exists h7'. assumption.
specialize (h2b h9). 
subst.
apply Extensionality_Ensembles.
red. split. 
red.
intros x h10.
destruct h10 as [y h10]. 
destruct h10 as [x h10 y h11]. subst.
destruct h10 as [h10].
destruct h10 as [h10l h10r].
assert (h11: (exist (Ensembles.In (Gen_Ens E)) (el_prod x) (el_prod_in_gen E x)) =  (exist (Ensembles.In (Gen_Ens E)) (el_prod x)
        (h3' (el_prod x)
           (Im_intro (Fin_map E signe mns) At
              [a : Fin_map E signe mns
              | Ensembles.In D
                  (exist (Ensembles.In (Gen_Ens E)) 
                     (el_prod a) (el_prod_in_gen E a)) /\ 
                el_prod a <> 0] el_prod x
              (intro_characteristic_sat
                 (fun a : Fin_map E signe mns =>
                  Ensembles.In D
                    (exist (Ensembles.In (Gen_Ens E)) 
                       (el_prod a) (el_prod_in_gen E a)) /\ 
                  el_prod a <> 0) x (conj h10l h10r)) 
              (el_prod x) eq_refl)))).
  apply proj1_sig_injective.
simpl. reflexivity.
unfold At in h11. unfold At.
rewrite <- h11.
assumption.
red.
intros x h11.
specialize (h6a _ h11). pose proof h6a as h12. clear h6a.
rewrite h2l in h12.
destruct h12 as [h12l h12r].
destruct h12r as [a h13].
assert (h14:Ensembles.In  (Im
             [a : Fin_map E signe mns
             | Ensembles.In D
                 (exist (Ensembles.In (Gen_Ens E)) 
                    (el_prod a) (el_prod_in_gen E a)) /\ 
               el_prod a <> 0] el_prod) (proj1_sig x)).
  rewrite h13.
  apply Im_intro with a.
  constructor. split.
  assert (h14: (exist (Ensembles.In (Gen_Ens E)) (el_prod a) (el_prod_in_gen E a)) = x).
  apply proj1_sig_injective. simpl. rewrite h13. reflexivity.
rewrite h14. assumption.
intro h14. rewrite h14 in h13. contradict h12l. apply proj1_sig_injective. simpl. assumption.
reflexivity.
pose proof (inc_sig_intro _ _ h3' _ h14) as h16.
assert (h15: (exist (Ensembles.In (Gen_Ens E)) (proj1_sig x)
            (h3' (proj1_sig x) h14)) = x).
apply proj1_sig_injective. simpl. reflexivity.
unfold At in h15. unfold At in h16.
rewrite h15 in h16.
assumption.
Qed.



End NormalForm.
Arguments el_prod [A] [E] _.
Arguments el_sum [A] [E] _.
Arguments eps [A] _ _.
Arguments eps' [A] _.   


Section Examples.
Variable fos:Field_of_Sets.
Definition psa_fos := psa (Xt fos).
Let Bt := Btype (Bc psa_fos).


Lemma fos_closed : @A_closed psa_fos (F fos).
constructor.
(* + *)
red.
intros A B.
destruct A.
destruct B. 
unfold Bplus_sub.
simpl.
apply (Union_closed fos); assumption.
(* * *)
red.
intros A B.
destruct A.
destruct B.
unfold Btimes_sub.
simpl.
apply (Int_closed fos); assumption.
(* 1 *)
red. simpl. apply full_in_F.
(* 0 *)
red. simpl. apply empty_in_F.
(* - *)
red.
intro A.
destruct A as [? h1].
unfold Bcomp_sub.
simpl.
apply (Comp_closed fos). assumption.
Qed.

Definition fos_sub_psa := @SubAlg psa_fos (F fos) fos_closed.

Variable Xt:Type.
Definition fin_cof_fos' := fin_cof_fos Xt.

Lemma fin_cof_closed : @A_closed (psa Xt) (F_fc Xt).
constructor.
(* + *)
red.
intros A B.
destruct A.
destruct B. 
unfold Bplus_sub.
simpl.
apply (Union_closed_fc); assumption.
(* * *)
red.
intros A B.
destruct A.
destruct B.
unfold Btimes_sub.
simpl.
apply (Int_closed_fc); assumption.
(* 1 *)
red. simpl. apply full_in_F_fc.
(* 0 *)
red. simpl. apply empty_in_F_fc.
(* - *)
red.
intro A.
destruct A.
unfold Bcomp_sub.
simpl.
apply (Comp_closed_fc); assumption.
Qed.

Definition fin_cof_sub_psa := @SubAlg (psa Xt) (F_fc Xt) fin_cof_closed. 

(* ". . . consider the field A of finite and cofinite subsets of
a set X. We shall prove that it is generated in P(X) by the set E of one-
element (that is, singleton) subsets of X." *)

Lemma fin_cof_gen_singletons' : @Gen (psa Xt) (SingF (Full_set Xt)) = fin_cof_sub_psa.
unfold fin_cof_sub_psa.
unfold Gen.
pose (@SA_cont_A (psa Xt) (SingF (Full_set Xt))) as F.
unfold SingF in F.
pose (int_saf_A F) as S.
simpl.
pose fin_cof_closed as h1.
pose proof (P_c _ _ h1) as h2.
assert (h3:Included (SingF (Full_set Xt)) (F_fc Xt)).
  red.
  intros A h4.
  inversion h4 as [x ? ? h5].
  rewrite h5.
  pose proof (Singleton_is_finite _ x) as h6.
  constructor. left; assumption.
assert (h4:int_saf_A (@SA_cont_A (psa Xt) (SingF (Full_set Xt))) = 
  (F_fc Xt)).
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red.
  intros sing h5.
  inversion h5 as [? h6].
  apply h6.
  constructor.
  split.
    (* left *)  
    red.
    intros sing'.
    unfold Bt in sing'.
    simpl in sing'.
    intro h7.
    inversion h7 as [y ? ? h8].
    rewrite h8.
    constructor.
    left.
    apply Singleton_is_finite.
    (* right *) assumption.
  (* >= *)
  red.
  intros S' h4.
  pose proof (union_singF_eq S') as h5.
  inversion h4 as [h6].
  assert (hfin : forall fin:(Btype (Bc (psa Xt))), Ensembles.In (F_fc Xt) fin -> Finite fin
    -> Ensembles.In S fin).
    intros fin h4' h6'.
    pose proof (union_singF_eq fin) as h5'. 
  induction h6' as [|fin' h7 h8 x h9].
  (* Empty *)
    constructor.
    intros E h6''.
    inversion h6'' as [h7].
    destruct h7 as [? h8].  
    pose proof (Z_c _ _ h8) as h9.
    red in h9.
    simpl in h9.
    assumption.
  (*Add*)
  unfold SingF in h5'.
  pose proof (Im_add _ _ fin' x (Singleton (U:=Xt))) as h10.
  unfold SingF.
  rewrite h10 in h5'.
  pose proof (family_union_add 
    (Im fin' (Singleton (U:=Xt))) (Singleton x)) as h11.
  rewrite h11 in h5'.
  rewrite h5'.
  pose proof (union_singF_eq fin') as h12.
  unfold SingF in h12.
  assert (h13:Ensembles.In (F_fc Xt) fin').
    constructor.
    left; assumption.
  pose proof (h8 h13 h12) as h14.
  rewrite <- h12.
  pose (Im (Full_set Xt) (Singleton (U:=Xt))) as J.
  assert (h15:Ensembles.In (int_saf_A (@SA_cont_A (psa Xt) J)) (Singleton x)).
    assert (h16:Ensembles.In J (Singleton x)).
      unfold J.
      apply Im_intro with x. apply Full_intro. trivial. 
      unfold SA_cont_A in F.
    pose (saf_A F) as EF.
    unfold Bt in EF.
    simpl in EF.
    unfold int_saf_A.
    assert (h17:forall (G:Family Xt), Ensembles.In EF G -> Included J G).
      intros G h18.
      inversion h18 as [h19].
      destruct h19 as [h19l].
      unfold J.
      assumption.
    constructor.
    unfold Bt. simpl.
    intros G h18.
    unfold EF in h17.
    pose proof (h17 G h18) as h19.
    auto with sets.

  pose proof (saf_closed F) as h20.    
  constructor.
  intros S0 h21.
  pose proof (h20 S0 h21) as h22.
  pose proof (P_c _ _ h22) as h23.
  red in h23.
  unfold SubBtype in h23.  unfold Bplus_sub in h23.
  unfold Bt in h23.
  simpl in h23.
  assert (h24: Ensembles.In S0 (Singleton x)).
    inversion h15 as [? h16].
    apply (h16 S0 h21).
  assert (h25: Ensembles.In S0 fin').
    inversion h14.
    apply (H S0 h21).
    apply (h23 (exist _ (Singleton x) h24) (exist _ fin' h25)).
destruct h6 as [h6l | h6r].
  (*h6l*)
  apply hfin; assumption.
  (*h6r*)
  assert (h7:Ensembles.In (F_fc Xt) (Complement S')).
    constructor.
    left; assumption.
  pose proof (hfin _ h7 h6r) as h8.
  pose proof (int_saf_A_closed F) as h9.
  pose proof (C_c _ _ h9) as h10.
  red in h10.
  unfold SubBtype in h10. unfold Bcomp_sub in h10.
  unfold Bt in h10. simpl in h10.
  pose proof (h10 (exist _ (Complement S') h8)) as h11.
  simpl in h11.
  pose proof (Complement_Complement _ S') as h12.
  rewrite <- h12.
  assumption. 
unfold Gen_Ens.
unfold SA_cont_A in h4. unfold int_saf_A in h4. simpl in h4.  
apply subalg_functional.
assumption.
Qed.


End Examples.
